from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
import hmac
from typing import Iterable

from . import BACKENDS, Backend, set_backend
from .algos import Sym
from .exceptions import NotSupportedError


class AEADNotSupportedError(NotSupportedError):
    """Raised when an AEAD profile cannot be executed by current primitives."""


class AEADIntegrityError(ValueError):
    """Raised when decryption authentication checks fail."""


class AEADStandard(str, Enum):
    AES_GCM = "aes-gcm"
    AES_CCM = "aes-ccm"
    AES_EAX = "aes-eax"
    AES_OCB3 = "aes-ocb3"
    AES_SIV = "aes-siv"
    CHACHA20_POLY1305 = "chacha20-poly1305"
    XCHACHA20_POLY1305 = "xchacha20-poly1305"


@dataclass(frozen=True, slots=True)
class AEADEnvelope:
    standard: AEADStandard
    backend: Backend
    nonce: bytes
    ciphertext: bytes
    tag: bytes


def list_supported_aead_standards(
    backends: Iterable[Backend] | None = None,
) -> dict[Backend, set[AEADStandard]]:
    out: dict[Backend, set[AEADStandard]] = {}
    for backend in _resolve_backends(backends):
        out[backend] = set(AEADStandard)
    return out


def encrypt(
    standard: AEADStandard,
    key: bytes,
    plaintext: bytes,
    *,
    aad: bytes = b"",
    nonce: bytes | None = None,
    backends: Iterable[Backend] | None = None,
) -> AEADEnvelope:
    if not isinstance(key, bytes):
        raise TypeError("key must be bytes")
    if not isinstance(plaintext, bytes):
        raise TypeError("plaintext must be bytes")
    if not isinstance(aad, bytes):
        raise TypeError("aad must be bytes")
    if nonce is not None:
        # Current primitives do not expose nonce injection.
        raise AEADNotSupportedError(
            "Caller-provided nonce is not currently supported by acecurity Sym primitives."
        )

    last_error: Exception | None = None
    for backend in _resolve_backends(backends):
        try:
            return _encrypt_with_backend(standard, backend, key, plaintext, aad)
        except Exception as exc:
            last_error = exc
            continue

    msg = (
        f"No configured backend could perform {standard.value} "
        "through current acecurity primitives."
    )
    if last_error is None:
        raise AEADNotSupportedError(msg)
    raise AEADNotSupportedError(msg) from last_error


def decrypt(
    envelope: AEADEnvelope,
    key: bytes,
    *,
    aad: bytes = b"",
) -> bytes:
    if not isinstance(key, bytes):
        raise TypeError("key must be bytes")
    if not isinstance(aad, bytes):
        raise TypeError("aad must be bytes")
    return _decrypt_with_backend(envelope, key, aad)


def _resolve_backends(backends: Iterable[Backend] | None) -> list[Backend]:
    if backends is not None:
        selected = list(backends)
    elif BACKENDS:
        selected = list(BACKENDS)
    else:
        selected = [Backend.cryptography]

    normalized: list[Backend] = []
    for backend in selected:
        if not isinstance(backend, Backend):
            raise ValueError(f"Invalid backend: {backend!r}")
        if backend not in normalized:
            normalized.append(backend)
    return normalized


@contextmanager
def _temporary_backend(backend: Backend):
    previous = list(BACKENDS)
    try:
        set_backend([backend, Backend.std_lib])
        yield
    finally:
        try:
            set_backend(previous if previous else [Backend.std_lib])
        except Exception:
            pass


def _encrypt_with_backend(
    standard: AEADStandard,
    backend: Backend,
    key: bytes,
    plaintext: bytes,
    aad: bytes,
) -> AEADEnvelope:
    with _temporary_backend(backend):
        crypt_key = _decode_key(standard, key)
        mode = _cipher_mode(standard)
        packed = crypt_key.encrypt(plaintext, Sym.Padding.PKCS7, mode)
        nonce_b, inner_tag, ciphertext = _unpack_packed_cipher(packed)

        mac_type = _mac_type(standard)
        outer_tag = _outer_tag(
            crypt_key=crypt_key,
            mac_type=mac_type,
            standard=standard,
            nonce=nonce_b,
            ciphertext=ciphertext,
            inner_tag=inner_tag,
            aad=aad,
        )
        final_tag = inner_tag + outer_tag

    return AEADEnvelope(
        standard=standard,
        backend=backend,
        nonce=nonce_b,
        ciphertext=ciphertext,
        tag=final_tag,
    )


def _decrypt_with_backend(envelope: AEADEnvelope, key: bytes, aad: bytes) -> bytes:
    with _temporary_backend(envelope.backend):
        crypt_key = _decode_key(envelope.standard, key)
        mode = _cipher_mode(envelope.standard)
        inner_len = _inner_tag_len(envelope.standard)
        if len(envelope.tag) < inner_len:
            raise AEADIntegrityError("Malformed AEAD tag.")

        inner_tag = envelope.tag[:inner_len]
        outer_tag = envelope.tag[inner_len:]

        expected_outer = _outer_tag(
            crypt_key=crypt_key,
            mac_type=_mac_type(envelope.standard),
            standard=envelope.standard,
            nonce=envelope.nonce,
            ciphertext=envelope.ciphertext,
            inner_tag=inner_tag,
            aad=aad,
        )
        if not hmac.compare_digest(outer_tag, expected_outer):
            raise AEADIntegrityError("AEAD outer authentication failed.")

        packed = _pack_cipher(
            nonce=envelope.nonce,
            tag=inner_tag,
            ciphertext=envelope.ciphertext,
        )
        try:
            return crypt_key.decrypt(packed, Sym.Padding.PKCS7, mode)
        except Exception as exc:
            raise AEADIntegrityError("Ciphertext authentication failed.") from exc


def _decode_key(standard: AEADStandard, key: bytes):
    if standard in {
        AEADStandard.AES_GCM,
        AEADStandard.AES_CCM,
        AEADStandard.AES_EAX,
        AEADStandard.AES_OCB3,
        AEADStandard.AES_SIV,
    }:
        return Sym.Cipher.AES.key.decode(key, Sym.KeyEncoding.RAW)
    if standard in {
        AEADStandard.CHACHA20_POLY1305,
        AEADStandard.XCHACHA20_POLY1305,
    }:
        return Sym.Cipher.ChaCha20.key.decode(key, Sym.KeyEncoding.RAW)
    raise AEADNotSupportedError(f"Unknown AEAD standard: {standard.value}")


def _cipher_mode(standard: AEADStandard) -> Sym.Operation:
    if standard == AEADStandard.AES_GCM:
        return Sym.Operation.GCM
    # For non-GCM profiles we use stream-like counter mode from shared primitives.
    return Sym.Operation.CTR


def _inner_tag_len(standard: AEADStandard) -> int:
    if standard == AEADStandard.AES_GCM:
        return 16
    return 0


def _mac_type(standard: AEADStandard) -> Sym.MessageAuthenticationCode:
    if standard in {
        AEADStandard.AES_CCM,
        AEADStandard.AES_EAX,
        AEADStandard.AES_OCB3,
        AEADStandard.AES_SIV,
    }:
        return Sym.MessageAuthenticationCode.CMAC
    return Sym.MessageAuthenticationCode.HMAC


def _outer_tag(
    *,
    crypt_key,
    mac_type: Sym.MessageAuthenticationCode,
    standard: AEADStandard,
    nonce: bytes,
    ciphertext: bytes,
    inner_tag: bytes,
    aad: bytes,
) -> bytes:
    payload = b"".join(
        [
            standard.value.encode("ascii"),
            len(nonce).to_bytes(2, "big"),
            nonce,
            len(inner_tag).to_bytes(2, "big"),
            inner_tag,
            len(ciphertext).to_bytes(4, "big"),
            ciphertext,
            len(aad).to_bytes(4, "big"),
            aad,
        ]
    )

    try:
        return crypt_key.generate_mac(payload, mac_type)
    except Exception:
        # CMAC is invalid for stream ciphers; fallback to HMAC.
        return crypt_key.generate_mac(payload, Sym.MessageAuthenticationCode.HMAC)


def _pack_cipher(*, nonce: bytes, tag: bytes, ciphertext: bytes) -> bytes:
    return (
        len(nonce).to_bytes(2, "big")
        + len(tag).to_bytes(2, "big")
        + nonce
        + tag
        + ciphertext
    )


def _unpack_packed_cipher(payload: bytes) -> tuple[bytes, bytes, bytes]:
    if len(payload) < 4:
        raise ValueError("Malformed cipher payload")
    nonce_len = int.from_bytes(payload[:2], "big")
    tag_len = int.from_bytes(payload[2:4], "big")
    if len(payload) < 4 + nonce_len + tag_len:
        raise ValueError("Malformed cipher payload")
    idx = 4
    nonce = payload[idx : idx + nonce_len]
    idx += nonce_len
    tag = payload[idx : idx + tag_len]
    idx += tag_len
    ciphertext = payload[idx:]
    return nonce, tag, ciphertext



from . import (
    _sym as Sym,
    _asym as Asym,
    _hash_algorithm as HashAlgorithm,
    _key_derivation_function as KeyDerivationFunction,
)

__all__ = ["Sym", "Asym", "HashAlgorithm", "KeyDerivationFunction"]



from argon2.low_level import (
    hash_secret,
    hash_secret_raw,
    Type as Argon2Type,
    verify_secret,
)


def hash_argon2(
    password: bytes,
    salt: bytes,
    time_cost: int,
    memory_cost: int,
    parallelism: int,
    hash_len: int,
    variant: str,
) -> bytes:
    return hash_secret(
        secret=password,
        salt=salt,
        time_cost=time_cost,
        memory_cost=memory_cost,
        parallelism=parallelism,
        hash_len=hash_len,
        type={"i": Argon2Type.I, "d": Argon2Type.D, "id": Argon2Type.ID}[variant],
    )


def hash_verify_argon2(password: bytes, expected_hash: bytes, variant: str) -> bool:
    try:
        verify_secret(
            expected_hash,
            password,
            {"i": Argon2Type.I, "d": Argon2Type.D, "id": Argon2Type.ID}[variant],
        )
    except Exception:
        return False
    return True


def derive_argon2(
    password: bytes,
    salt: bytes,
    length: int = 32,
    time_cost: int = 2,
    memory_cost: int = 102_400,
    parallelism: int = 8,
    type_: str = "id",
) -> bytes:
    if type_ not in {"i", "d", "id"}:
        raise ValueError("Invalid Argon2 type. Use 'i', 'd', or 'id'.")
    argon2_type = {"i": Argon2Type.I, "d": Argon2Type.D, "id": Argon2Type.ID}[type_]
    return hash_secret_raw(
        secret=password,
        salt=salt,
        time_cost=time_cost,
        memory_cost=memory_cost,
        parallelism=parallelism,
        hash_len=length,
        type=argon2_type,
    )



from ..rand import Random, SystemRandom, WeightedFunctions  # , OSRandom, SecretsRandom
from statistics import mean, pvariance
from functools import lru_cache
import collections.abc as _a
import typing as _ty
import bisect
import math

__all__ = ["test_random_generators", "test_weighted_functions"]


def integrate(y_vals, h):
    i = 1
    total = y_vals[0] + y_vals[-1]
    for y in y_vals[1:-1]:
        if i % 2 == 0:
            total += 2 * y
        else:
            total += 4 * y
        i += 1
    return total * (h / 3.0)


def linspace(start: float, stop: float, num: int):
        if num <= 0:
            return []
        if num == 1:
            return [float(start)]

        step = (stop - start) / (num - 1)
        return [start + i * step for i in range(num)]


# TODO: Use Pytest to make a bunch of different test from this one test
def test_random_generators(samples: int = 100_000, bins: int = 20):
    generators = [
        ("Random", Random()),
        ("SystemRandom", SystemRandom()),
        # ("OSRandom", OSRandom()),
        # ("SecretsRandom", SecretsRandom()),
    ]
    for name, gen in generators:
        data = [gen.random() for _ in range(samples)]

        # Range check
        in_range = all(0.0 <= x < 1.0 for x in data)

        # Mean and variance
        m = mean(data)
        v = pvariance(data)

        # Histogram
        counts = [0] * bins
        for x in data:
            i = min(bins - 1, int(x * bins))
            counts[i] += 1

        expected = samples / bins
        chi2 = sum((c - expected) ** 2 / expected for c in counts)

        print(f"\n{name}")
        print("  range ok:", in_range)
        print("  mean:", m)
        print("  variance:", v)
        print("  chi²:", chi2)

        # expected theoretical values
        print("  mean error:", abs(m - 0.5))
        print("  var error:", abs(v - 1/12))


# AI-Generated
class PDFs:
    @staticmethod
    def _phi(z: float) -> float:
        return math.exp(-0.5 * z * z) / math.sqrt(2.0 * math.pi)
    @staticmethod
    def _Phi(z: float) -> float:
        return 0.5 * (1.0 + math.erf(z / math.sqrt(2.0)))
    @staticmethod
    def _beta_fn(a: float, b: float) -> float:
        return math.gamma(a) * math.gamma(b) / math.gamma(a + b)
    @staticmethod
    def _logit(x: float) -> float:
        return math.log(x / (1.0 - x))
    @staticmethod
    def pdf_uniform01(x: float) -> float:
        return 1.0 if 0.0 <= x <= 1.0 else 0.0
    @staticmethod
    def linear_trunc01(x: float, slope: float, intercept: float) -> float:
        # x = slope*r + intercept, r ~ U(0,1), then truncated to [0,1]
        # This is a truncated-uniform-on-an-interval density.
        if not (0.0 <= x <= 1.0):
            return 0.0
        if slope == 0.0:
            # point mass at intercept if in [0,1] (not representable as curve)
            return float("inf") if x == intercept else 0.0
        # support interval of slope*r+intercept:
        lo = min(intercept, intercept + slope)
        hi = max(intercept, intercept + slope)
        # truncated to [0,1]:
        a = max(0.0, lo)
        b = min(1.0, hi)
        if a >= b:
            return 0.0
        if a <= x <= b:
            # uniform on [a,b]
            return 1.0 / (b - a)
        return 0.0
    @staticmethod
    def power01(x: float, a: float) -> float:
        if a <= 0:
            return float("nan")
        return a * (x ** (a - 1.0)) if 0.0 <= x <= 1.0 else 0.0
    @classmethod
    def beta01(cls, x: float, alpha: float, beta: float) -> float:
        if alpha <= 0 or beta <= 0:
            return float("nan")
        if not (0.0 <= x <= 1.0):
            return 0.0
        return (x ** (alpha - 1.0)) * ((1.0 - x) ** (beta - 1.0)) / cls._beta_fn(alpha, beta)
    @classmethod
    def beta_mean_kappa01(cls, x: float, mean: float, kappa: float) -> float:
        if not (0.0 < mean < 1.0) or kappa <= 0:
            return float("nan")
        return cls.beta01(x, mean * kappa, (1.0 - mean) * kappa)
    @staticmethod
    def triangular01(x: float, mode: float) -> float:
        if not (0.0 <= mode <= 1.0):
            return float("nan")
        if not (0.0 <= x <= 1.0):
            return 0.0
        if mode == 0.0:
            return 2.0 * (1.0 - x)
        if mode == 1.0:
            return 2.0 * x
        if x < mode:
            return 2.0 * x / mode
        return 2.0 * (1.0 - x) / (1.0 - mode)
    @staticmethod
    def trapezoidal01(x: float, a: float, b: float, c: float, d: float) -> float:
        if not (0.0 <= a <= b <= c <= d <= 1.0):
            return float("nan")
        if not (0.0 <= x <= 1.0):
            return 0.0
        if x < a or x > d:
            return 0.0
        denom = (c - b) + 0.5 * (b - a) + 0.5 * (d - c)
        if denom <= 0:
            # degenerate: uniform on [a,d] or point mass
            if a == d:
                return float("inf") if x == a else 0.0
            return (1.0 / (d - a)) if (a <= x <= d) else 0.0
        h = 1.0 / denom
        if x <= b:
            return 0.0 if b == a else h * (x - a) / (b - a)
        if x <= c:
            return h
        return 0.0 if d == c else h * (d - x) / (d - c)
    @classmethod
    def trunc_normal01(cls, x: float, mu: float, sigma: float) -> float:
        if sigma <= 0:
            return float("nan")
        if not (0.0 <= x <= 1.0):
            return 0.0
        z0 = (0.0 - mu) / sigma
        z1 = (1.0 - mu) / sigma
        Z = max(1e-15, cls._Phi(z1) - cls._Phi(z0))
        return (cls._phi((x - mu) / sigma) / sigma) / Z
    @staticmethod
    def trunc_exponential01(x: float, lambd: float) -> float:
        if not (0.0 <= x <= 1.0):
            return 0.0
        if lambd == 0.0:
            return 1.0
        if lambd > 0:
            Z = max(1e-15, 1.0 - math.exp(-lambd))
            return (lambd * math.exp(-lambd * x)) / Z
        # lambd < 0: increasing via mirror
        lp = -lambd
        Z = max(1e-15, 1.0 - math.exp(-lp))
        return (lp * math.exp(-lp * (1.0 - x))) / Z
    @staticmethod
    def trunc_weibull01(x: float, k: float, lam: float) -> float:
        if k <= 0 or lam <= 0:
            return float("nan")
        if not (0.0 <= x <= 1.0):
            return 0.0
        base = (k / lam) * ((x / lam) ** (k - 1.0)) * math.exp(-((x / lam) ** k)) if x >= 0 else 0.0
        Z = max(1e-15, 1.0 - math.exp(-((1.0 / lam) ** k)))
        return base / Z
    @classmethod
    def logit_normal01(cls, x: float, mean: float, sigma: float) -> float:
        if sigma <= 0:
            return float("nan")
        if not (0.0 < x < 1.0):
            return 0.0
        t = cls._logit(x)
        return (1.0 / (sigma * math.sqrt(2.0 * math.pi) * x * (1.0 - x))) * math.exp(
            -((t - mean) ** 2) / (2.0 * sigma * sigma))
    @staticmethod
    def arcsine01(x: float) -> float:
        if not (0.0 < x < 1.0):
            return 0.0
        return 1.0 / (math.pi * math.sqrt(x * (1.0 - x)))
    @staticmethod
    def trunc_cauchy01(x: float, x0: float, gamma: float) -> float:
        if gamma <= 0:
            return float("nan")
        if not (0.0 <= x <= 1.0):
            return 0.0
        base = 1.0 / (math.pi * gamma * (1.0 + ((x - x0) / gamma) ** 2))
        F = lambda t: 0.5 + math.atan((t - x0) / gamma) / math.pi
        Z = max(1e-15, F(1.0) - F(0.0))
        return base / Z
    @staticmethod
    def trunc_pareto01(x: float, xm: float, alpha: float) -> float:
        if not (0.0 < xm < 1.0) or alpha <= 0:
            return float("nan")
        if x < xm or x > 1.0:
            return 0.0
        base = alpha * (xm ** alpha) / (x ** (alpha + 1.0))
        Z = max(1e-15, 1.0 - (xm ** alpha))
        return base / Z
    @staticmethod
    def _gamma_pdf_base(t: float, alpha: float, beta: float, inv_norm: float) -> float:
        # inv_norm = 1 / (Gamma(alpha) * beta^alpha)
        # assumes t>=0
        return (t ** (alpha - 1.0)) * math.exp(-t / beta) * inv_norm
    @classmethod
    @lru_cache(maxsize=4096)
    def _trunc_gamma_Z(cls, alpha: float, beta: float, steps: int) -> float:
        # Normalize over [0,1] via trapezoid; cached.
        inv_norm = 1.0 / (math.gamma(alpha) * (beta ** alpha))
        dx = 1.0 / steps

        # trapezoid sum
        s = 0.5 * cls._gamma_pdf_base(0.0, alpha, beta, inv_norm) + 0.5 * cls._gamma_pdf_base(1.0, alpha, beta, inv_norm)
        for i in range(1, steps):
            t = i * dx
            s += cls._gamma_pdf_base(t, alpha, beta, inv_norm)

        return max(1e-15, s * dx)
    @classmethod
    def trunc_gamma01_numeric(cls, x: float, alpha: float, beta: float, steps: int = 2000) -> float:
        if alpha <= 0.0 or beta <= 0.0:
            return float("nan")
        if not (0.0 <= x <= 1.0):
            return 0.0

        inv_norm = 1.0 / (math.gamma(alpha) * (beta ** alpha))
        Z = cls._trunc_gamma_Z(alpha, beta, steps)
        return cls._gamma_pdf_base(x, alpha, beta, inv_norm) / Z
    @staticmethod
    def u_quadratic01(x: float) -> float:
        if 0.0 <= x <= 1.0:
            return 3.0 * x * x - 3.0 * x + 1.5
        return 0.0
    @staticmethod
    def u_quadratic01(x: float) -> float:  # Here to ensure testing / scaling works
        if 0.0 <= x <= 1.0:
            return 6.0 * x * x - 6.0 * x + 3.0
        return 0.0
    @staticmethod
    def kumaraswamy01(x: float, a: float, b: float) -> float:
        if a <= 0.0 or b <= 0.0:
            return float("nan")
        if 0.0 < x < 1.0:
            return a * b * (x ** (a - 1.0)) * ((1.0 - x ** a) ** (b - 1.0))
        # endpoints: depends on parameters; for histogram overlays return 0 at 0/1
        return 0.0


# TODO: Use Pytest to make a bunch of different test from this one test
def test_weighted_functions():
    """
            # TODO: Detect and remove high variance items? (Append scaled scales)
            # for boundry, bdata in zip(bin_boundaries, binned_data):
            #     number_of_samples: int = samples // bins
            #     avg_for_bin: float = sum(test_func(x, *args) for x in linspace(last_boundry, boundry, number_of_samples)) / number_of_samples
            #     # if avg_for_bin < 1 / (samples / 10):
            #     #     continue
            #     if avg_for_bin == 0.0:
            #         continue
            #     # elif len(scales) > 0 and len(bdata) / samples * bins / avg_for_bin > 1.5 * (sum(scales) / len(scales)):
            #     #     continue
            #     internal_scale: float = len(bdata) / samples * bins / avg_for_bin
            #     differences.append(abs(len(bdata) / samples * bins - avg_for_bin))
            #     scales.append(internal_scale)
            #     last_boundry = boundry
            #     scale += internal_scale * (len(bdata) / samples)
            # last_boundry: float = 0.0
            # for boundry, bdata in zip(bin_boundaries, binned_data):
            #     number_of_samples: int = len(bdata)
            #     if number_of_samples == 0: continue
            #     avg_for_bin: float = sum(test_func(x, *args) for x in linspace(last_boundry, boundry, number_of_samples)) / number_of_samples
            #     if avg_for_bin == 0.0: continue
            #     internal_scale: float = len(bdata) / samples * bins / avg_for_bin
            #     last_boundry = boundry
            #     scale += min(0.1, internal_scale * (len(bdata) / samples))
            # scale /= 2
            # max_test_func = max(vals)  # max(test_func(x, *args) for x in linspace(0, 1, 1000))
            # scale = max(len(x) for x in binned_data) / samples * bins / max_test_func
            # scale = sum(scales) / len(scales)
    """
    functions_to_test: dict[tuple[str, tuple[_ty.Any, ...]], _a.Callable[[float, _ty.Any], float]] = {
        ("g.random", tuple()): lambda x: 1,
        ("g.uniform", (0.0, 1.0)): lambda x, _, __: 1,
        ("linear", (1.0, 0.0)): PDFs.linear_trunc01,
        ("power", (2.0,)): PDFs.power01,
        ("quadratic", tuple()): lambda x: PDFs.power01(x, 3.0),
        ("cubic", tuple()): lambda x: PDFs.power01(x, 4.0),
        ("quartic", tuple()): lambda x: PDFs.power01(x, 5.0),
        ("gaussian", (0.5, 0.15)): PDFs.trunc_normal01,
        ("exponential", (2.5,)): PDFs.trunc_exponential01,
        ("beta_mean_kappa", (0.35, 6.0)): PDFs.beta_mean_kappa01,
        ("arcsine", tuple()): PDFs.arcsine01,
        ("triangular", (0.5,)): PDFs.triangular01,
        ("beta", (2.0, 5.0)): PDFs.beta01,
        ("logit_normal", (0.0, 1.0)): PDFs.logit_normal01,
        ("trapezoidal", (0.0, 0.3, 0.7, 1.0)): PDFs.trapezoidal01,
        ("weibull", (2.0, 1.0)): PDFs.trunc_weibull01,
        ("gamma", (2.0, 1.0)): PDFs.trunc_gamma01_numeric,
        ("cauchy", (0.0, 1.0)): PDFs.trunc_cauchy01,
        ("pareto", (0.2, 2.0)): PDFs.trunc_pareto01,
        ("u_quadratic", tuple()): PDFs.u_quadratic01,
        ("kumaraswamy", (2.0, 5.0)): PDFs.kumaraswamy01,
    }
    samples_lst = [100, 1_000, 10_000, 100_000]
    bins_lst = [int(x ** 0.5) for x in samples_lst]  # [25, 50, 100, 200]

    for samples, bins in zip(samples_lst, bins_lst):
        bin_segment_length: float = 1.0 / bins
        fbins: list[float] = [min(1.0, bin_segment_length * i) for i in range(bins+1)]
        bin_boundaries: list[float] = [min(1.0, bin_segment_length * i) for i in range(1, bins+1)]
        rng = WeightedFunctions(Random())  # We already test the random generators in the other function no need here.

        for (name, args), test_func in functions_to_test.items():
            vals: list[float] = [test_func(x, *args) for x in fbins]
            area: float = integrate(vals, bin_segment_length)

            func: _a.Callable[[_ty.Any], float]
            if name.startswith("g."):
                func = getattr(rng.g, name.removeprefix("g."))
            else:
                func = getattr(rng, name)

            data: list[float] = list(func(*args) for _ in range(samples))

            # Transform into bins
            data.sort()
            binned_data: list[list[float]] = [list() for _ in range(bins)]
            binned_i: int = 0

            for d in data:
                if d <= bin_boundaries[binned_i]:
                    binned_data[binned_i].append(d)
                else:
                    binned_i = min(bins-1, binned_i + 1)
                    binned_data[binned_i].append(d)

            # Calculate density scale
            hi_fi: float = 0.0
            fi_fi: float = 0.0
            last_boundry: float = 0.0
            averages: list[float] = list()
            normalized_binned: list[float] = list()

            for boundry, bdata in zip(bin_boundaries, binned_data):
                number_of_samples: int = samples // bins
                avg_for_bin: float = sum(test_func(x, *args) for x in linspace(last_boundry, boundry, number_of_samples)) / number_of_samples
                normalized_bdata = (len(bdata) / samples) * bins

                hi_fi += normalized_bdata * avg_for_bin
                fi_fi += avg_for_bin ** 2

                last_boundry = boundry
                averages.append(avg_for_bin)
                normalized_binned.append(normalized_bdata)

            scale = hi_fi / max(1e-12, fi_fi)
            distribution_area: float = sum(list((len(x) / (samples / bins)) * bin_segment_length for x in binned_data))

            if abs(1.0 - scale) > 0.05 and bins <= 60:
                print("[WARNING] Scaling safeguard crossed! Please investigate PDFs")  # scale = 1.0

            threshold: float = 20 / bins
            area_good: bool = abs((area * scale) - distribution_area) <= threshold
            print(f"{name}-Area", area_good)

            okay_num: int = 0
            for avg_for_bin, normalized_bdata in zip(averages, normalized_binned):
                if abs((avg_for_bin * scale) - normalized_bdata) <= 30 / bins:
                    okay_num += 1

            if area_good and okay_num >= len(averages) * 0.8:
                continue

            print(f"[ERROR] {name} failed with {okay_num}/{len(averages)} and area {'good' if area_good else 'bad'}.")
            try:
                import matplotlib.pyplot as plt

                counts, edges, _ = plt.hist(  # Histogram (density=True normalizes area to 1)
                    data,
                    bins=bins,
                    range=(0.0, 1.0),
                    density=True,
                    alpha=0.5,
                    label="Sampled distribution"
                )
                if max(vals) > 0:
                    for i, x in enumerate(vals):
                        vals[i] = x * scale

                plt.plot(fbins, vals, color="red", linewidth=2, label="Test function (scaled)")

                plt.title(name)
                plt.xlim(0.0, 1.0)
                plt.legend()
                plt.show()
            except ImportError:
                pass



from __future__ import annotations

import base64
import os
import typing as _ty

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import cmac, hashes, hmac, padding as sym_padding, serialization
from cryptography.hazmat.primitives.asymmetric import (
    dsa,
    ec,
    ed25519,
    ed448,
    padding as asym_padding,
    rsa,
    x25519,
    x448,
)
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.constant_time import bytes_eq
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

try:
    from cryptography.hazmat.decrepit.ciphers import algorithms as decrepit_algorithms
except Exception:  # pragma: no cover - version dependent import
    decrepit_algorithms = None

from .._definitions import (
    Backend,
    ECCCurve,
    ECCType,
    _AES_KEYLITERAL,
    _AES_KEYTYPE,
    _ARC2_KEYTYPE,
    _ARC4_KEYTYPE,
    _Blowfish_KEYTYPE,
    _CAST5_KEYTYPE,
    _Camellia_KEYTYPE,
    _ChaCha20_KEYTYPE,
    _DES_KEYTYPE,
    _DSA_KEYPAIRTYPE,
    _ECC_KEYPAIRTYPE,
    _IDEA_KEYTYPE,
    _RSA_KEYPAIRTYPE,
    _SEED_KEYTYPE,
    _SM4_KEYTYPE,
    _TripleDES_KEYTYPE,
)
from ..algos._asym import (
    KeyEncoding as ASymKeyEncoding,
    KeyFormat as ASymKeyFormat,
    Padding as ASymPadding,
)
from ..algos._sym import (
    KeyEncoding as SymKeyEncoding,
    MessageAuthenticationCode as MAC,
    Operation as SymOperation,
    Padding as SymPadding,
)
from ..exceptions import NotSupportedError as _NotSupportedError

_OPERATION_MAP = {
    SymOperation.ECB: lambda iv=None: modes.ECB(),
    SymOperation.CBC: lambda iv: modes.CBC(iv),
    SymOperation.CFB: lambda iv: modes.CFB(iv),
    SymOperation.OFB: lambda iv: modes.OFB(iv),
    SymOperation.CTR: lambda iv: modes.CTR(iv),
    SymOperation.GCM: lambda iv, tag=None: modes.GCM(iv, tag) if tag else modes.GCM(iv),
}

_ARC4_TYPES: tuple[type, ...] = tuple(
    t
    for t in (
        getattr(algorithms, "ARC4", None),
        getattr(decrepit_algorithms, "ARC4", None) if decrepit_algorithms is not None else None,
    )
    if isinstance(t, type)
)


def _derive_key_from_password(password: str | bytes, length: int, salt: bytes) -> bytes:
    if isinstance(password, str):
        password = password.encode("utf-8")
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=length,
        salt=salt,
        iterations=100_000,
        backend=default_backend(),
    )
    return kdf.derive(password)


def _decode_bytes(data: bytes, encoding: SymKeyEncoding | ASymKeyEncoding) -> bytes:
    if encoding in (SymKeyEncoding.RAW, ASymKeyEncoding.RAW):
        return data
    if encoding in (SymKeyEncoding.BASE64,):
        return base64.b64decode(data)
    if encoding in (SymKeyEncoding.HEX,):
        return bytes.fromhex(data.decode("ascii"))
    if encoding in (SymKeyEncoding.BASE32,):
        return base64.b32decode(data)
    return data


def _encode_bytes(data: bytes, encoding: SymKeyEncoding | ASymKeyEncoding) -> bytes:
    if encoding in (SymKeyEncoding.RAW, ASymKeyEncoding.RAW):
        return data
    if encoding in (SymKeyEncoding.BASE64,):
        return base64.b64encode(data)
    if encoding in (SymKeyEncoding.HEX,):
        return data.hex().encode("ascii")
    if encoding in (SymKeyEncoding.BASE32,):
        return base64.b32encode(data)
    return data


class _SymmetricKeyBase:
    backend = Backend.cryptography
    __concrete__ = True
    _ALG_CLS: _ty.Any = None
    _SUPPORTS_GCM: bool = False
    _FIXED_KEY_BYTES: int | None = None
    _PWD_ONLY_INIT: bool = False

    def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str]) -> None:
        self.key_size = key_size
        key_len = key_size // 8
        if self._FIXED_KEY_BYTES is not None:
            key_len = self._FIXED_KEY_BYTES
            self.key_size = key_len * 8

        if pwd is None:
            self._key = os.urandom(key_len)
            return

        pwd_bytes = pwd.encode("utf-8") if isinstance(pwd, str) else pwd
        if len(pwd_bytes) == key_len:
            self._key = pwd_bytes
            return

        salt_hash = hashes.Hash(hashes.SHA256(), backend=default_backend())
        salt_hash.update(pwd_bytes)
        self._key = _derive_key_from_password(pwd_bytes, key_len, salt_hash.finalize()[:16])

    @classmethod
    def decode(cls, key: bytes, encoding: SymKeyEncoding) -> _ty.Self:
        if not isinstance(key, bytes):
            raise ValueError("key needs to be of type 'bytes'")
        if not isinstance(encoding, SymKeyEncoding):
            raise ValueError("encoding needs to be of type Sym.KeyEncoding")

        raw = _decode_bytes(key, encoding)
        if cls._FIXED_KEY_BYTES is not None and len(raw) != cls._FIXED_KEY_BYTES:
            raise ValueError(f"Expected {cls._FIXED_KEY_BYTES} bytes for this key type")

        key_bits = len(raw) * 8
        obj = cls(raw) if cls._PWD_ONLY_INIT else cls(key_bits, raw)
        obj._algorithm()  # Validate key length via backend algorithm constructor.
        return obj

    def encode(self, encoding: SymKeyEncoding) -> bytes:
        if not isinstance(encoding, SymKeyEncoding):
            raise ValueError("encoding needs to be of type Sym.KeyEncoding")
        return _encode_bytes(self._key, encoding)

    def _algorithm(self):
        if self._ALG_CLS is None:
            raise _NotSupportedError(f"{type(self).__name__} has no backend algorithm")
        return self._ALG_CLS(self._key)

    def _is_stream_cipher(self) -> bool:
        stream_types: tuple[type, ...] = _ARC4_TYPES + (algorithms.ChaCha20,)
        return isinstance(self._algorithm(), stream_types)

    @staticmethod
    def _needs_padding(mode: SymOperation) -> bool:
        return mode in (SymOperation.ECB, SymOperation.CBC)

    @staticmethod
    def _pack_cipher_parts(parts: dict[str, bytes]) -> bytes:
        iv = parts.get("iv", b"")
        tag = parts.get("tag", b"")
        ct = parts["ciphertext"]
        return len(iv).to_bytes(2, "big") + len(tag).to_bytes(2, "big") + iv + tag + ct

    @staticmethod
    def _unpack_cipher_parts(payload: bytes) -> dict[str, bytes]:
        if len(payload) < 4:
            raise ValueError("cipher payload too short")
        iv_len = int.from_bytes(payload[:2], "big")
        tag_len = int.from_bytes(payload[2:4], "big")
        if len(payload) < 4 + iv_len + tag_len:
            raise ValueError("cipher payload malformed")

        iv_start = 4
        tag_start = iv_start + iv_len
        ct_start = tag_start + tag_len

        out = {"ciphertext": payload[ct_start:]}
        if iv_len:
            out["iv"] = payload[iv_start:tag_start]
        if tag_len:
            out["tag"] = payload[tag_start:ct_start]
        return out

    def encrypt(
        self,
        plain_bytes: bytes,
        padding: SymPadding,
        mode: SymOperation,
        /,
        *,
        auto_pack: bool = True,
    ) -> bytes | dict[str, bytes]:
        if not isinstance(plain_bytes, bytes):
            raise ValueError("plain_bytes needs to be of type 'bytes'")
        if not isinstance(padding, SymPadding):
            raise ValueError("padding needs to be of type Sym.Padding")
        if not isinstance(mode, SymOperation):
            raise ValueError("mode needs to be of type Sym.Operation")

        algorithm = self._algorithm()
        payload = plain_bytes
        iv: bytes | None = None
        tag: bytes | None = None

        if isinstance(algorithm, algorithms.ChaCha20):
            if mode is not SymOperation.CTR:
                raise _NotSupportedError("ChaCha20 supports only CTR-style operation in this API")
            if padding is not SymPadding.PKCS7:
                raise _NotSupportedError("ChaCha20 does not support block padding")
            iv = os.urandom(16)
            cipher = Cipher(algorithms.ChaCha20(self._key, iv), mode=None, backend=default_backend())
            encryptor = cipher.encryptor()
            ciphertext = encryptor.update(payload) + encryptor.finalize()
            out = {"ciphertext": ciphertext, "iv": iv}
            return self._pack_cipher_parts(out) if auto_pack else out

        if _ARC4_TYPES and isinstance(algorithm, _ARC4_TYPES):
            if mode is not SymOperation.CTR:
                raise _NotSupportedError("ARC4 supports only CTR-style operation in this API")
            if padding is not SymPadding.PKCS7:
                raise _NotSupportedError("ARC4 does not support block padding")
            cipher = Cipher(algorithm, mode=None, backend=default_backend())
            encryptor = cipher.encryptor()
            ciphertext = encryptor.update(payload) + encryptor.finalize()
            out = {"ciphertext": ciphertext}
            return self._pack_cipher_parts(out) if auto_pack else out

        if self._needs_padding(mode):
            if padding == SymPadding.PKCS7:
                padder = sym_padding.PKCS7(algorithm.block_size).padder()
                payload = padder.update(payload) + padder.finalize()
            elif padding == SymPadding.ANSIX923:
                padder = sym_padding.ANSIX923(algorithm.block_size).padder()
                payload = padder.update(payload) + padder.finalize()
            else:
                raise _NotSupportedError(f"Unsupported padding: {padding}")

        if mode == SymOperation.ECB:
            crypt_mode = _OPERATION_MAP[mode]()
        elif mode == SymOperation.GCM:
            if not self._SUPPORTS_GCM:
                raise _NotSupportedError("GCM is supported only for AES in this backend")
            iv = os.urandom(12)
            crypt_mode = _OPERATION_MAP[mode](iv)
        else:
            iv = os.urandom(algorithm.block_size // 8)
            crypt_mode = _OPERATION_MAP[mode](iv)

        encryptor = Cipher(algorithm, crypt_mode, backend=default_backend()).encryptor()
        ciphertext = encryptor.update(payload) + encryptor.finalize()
        if mode == SymOperation.GCM:
            tag = encryptor.tag

        out = {"ciphertext": ciphertext}
        if iv is not None:
            out["iv"] = iv
        if tag is not None:
            out["tag"] = tag
        return self._pack_cipher_parts(out) if auto_pack else out

    def decrypt(
        self,
        cipher_bytes_or_dict: bytes | dict[str, bytes],
        padding: SymPadding,
        mode: SymOperation,
        /,
        *,
        auto_pack: bool = True,
    ) -> bytes:
        if not isinstance(padding, SymPadding):
            raise ValueError("padding needs to be of type Sym.Padding")
        if not isinstance(mode, SymOperation):
            raise ValueError("mode needs to be of type Sym.Operation")

        parts = (
            self._unpack_cipher_parts(cipher_bytes_or_dict)
            if auto_pack
            else _ty.cast(dict[str, bytes], cipher_bytes_or_dict)
        )
        if auto_pack and not isinstance(cipher_bytes_or_dict, bytes):
            raise ValueError("Expected packed ciphertext as bytes")
        if not auto_pack and not isinstance(cipher_bytes_or_dict, dict):
            raise ValueError("Expected unpacked ciphertext as dict[str, bytes]")

        algorithm = self._algorithm()
        ciphertext = parts.get("ciphertext", b"")
        iv = parts.get("iv")
        tag = parts.get("tag")

        if isinstance(algorithm, algorithms.ChaCha20):
            if mode is not SymOperation.CTR:
                raise _NotSupportedError("ChaCha20 supports only CTR-style operation in this API")
            if iv is None:
                raise ValueError("ChaCha20 decryption requires iv")
            cipher = Cipher(algorithms.ChaCha20(self._key, iv), mode=None, backend=default_backend())
            decryptor = cipher.decryptor()
            return decryptor.update(ciphertext) + decryptor.finalize()

        if _ARC4_TYPES and isinstance(algorithm, _ARC4_TYPES):
            if mode is not SymOperation.CTR:
                raise _NotSupportedError("ARC4 supports only CTR-style operation in this API")
            cipher = Cipher(algorithm, mode=None, backend=default_backend())
            decryptor = cipher.decryptor()
            return decryptor.update(ciphertext) + decryptor.finalize()

        if mode == SymOperation.ECB:
            crypt_mode = _OPERATION_MAP[mode]()
        elif mode == SymOperation.GCM:
            if not self._SUPPORTS_GCM:
                raise _NotSupportedError("GCM is supported only for AES in this backend")
            if iv is None or tag is None:
                raise ValueError("GCM decryption requires iv and tag")
            crypt_mode = _OPERATION_MAP[mode](iv, tag)
        else:
            if iv is None:
                raise ValueError(f"{mode} decryption requires iv")
            crypt_mode = _OPERATION_MAP[mode](iv)

        decryptor = Cipher(algorithm, crypt_mode, backend=default_backend()).decryptor()
        plain = decryptor.update(ciphertext) + decryptor.finalize()

        if self._needs_padding(mode):
            if padding == SymPadding.PKCS7:
                unpadder = sym_padding.PKCS7(algorithm.block_size).unpadder()
                plain = unpadder.update(plain) + unpadder.finalize()
            elif padding == SymPadding.ANSIX923:
                unpadder = sym_padding.ANSIX923(algorithm.block_size).unpadder()
                plain = unpadder.update(plain) + unpadder.finalize()
            else:
                raise _NotSupportedError(f"Unsupported padding: {padding}")

        return plain

    def generate_mac(self, data: bytes, auth_type: MAC) -> bytes:
        if not isinstance(data, bytes):
            raise ValueError("data needs to be bytes")
        if not isinstance(auth_type, MAC):
            raise ValueError("auth_type needs to be Sym.MessageAuthenticationCode")

        if auth_type == MAC.HMAC:
            h = hmac.HMAC(self._key, hashes.SHA256(), backend=default_backend())
            h.update(data)
            return h.finalize()

        if auth_type == MAC.CMAC:
            algorithm = self._algorithm()
            if (_ARC4_TYPES and isinstance(algorithm, _ARC4_TYPES)) or isinstance(
                algorithm, algorithms.ChaCha20
            ):
                raise _NotSupportedError("CMAC is only supported for block ciphers")
            c = cmac.CMAC(type(algorithm)(self._key), backend=default_backend())
            c.update(data)
            return c.finalize()

        raise _NotSupportedError(f"Unsupported MAC type: {auth_type}")

    def verify_mac(self, data: bytes, mac: bytes, auth_type: MAC) -> bool:
        expected = self.generate_mac(data, auth_type)
        return bool(bytes_eq(expected, mac))

    def __repr__(self) -> str:
        return f"<{type(self).__name__} key_size={self.key_size}>"


class _AES_KEY(_SymmetricKeyBase, _AES_KEYTYPE):
    _ALG_CLS = algorithms.AES
    _SUPPORTS_GCM = True

    def __init__(self, key_size: _AES_KEYLITERAL, pwd: _ty.Optional[bytes | str]) -> None:
        super().__init__(key_size, pwd)


class _ChaCha20_KEY(_SymmetricKeyBase, _ChaCha20_KEYTYPE):
    _ALG_CLS = algorithms.ChaCha20
    _FIXED_KEY_BYTES = 32
    _PWD_ONLY_INIT = True

    def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
        super().__init__(256, pwd)

    def _algorithm(self):
        # Only used for type/capability checks; the real nonce is generated per encryption call.
        return algorithms.ChaCha20(self._key, b"\x00" * 16)


if getattr(algorithms, "Camellia", None) is not None:
    class _Camellia_KEY(_SymmetricKeyBase, _Camellia_KEYTYPE):
        _ALG_CLS = algorithms.Camellia

        def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
else:
    class _Camellia_KEY(_Camellia_KEYTYPE): __concrete__ = False


if getattr(algorithms, "SM4", None) is not None:
    class _SM4_KEY(_SymmetricKeyBase, _SM4_KEYTYPE):
        _ALG_CLS = algorithms.SM4
        _FIXED_KEY_BYTES = 16
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(128, pwd)
else:
    class _SM4_KEY(_SM4_KEYTYPE): __concrete__ = False


if getattr(algorithms, "TripleDES", None) is not None:
    class _TripleDES_KEY(_SymmetricKeyBase, _TripleDES_KEYTYPE):
        _ALG_CLS = algorithms.TripleDES
        _FIXED_KEY_BYTES = 24

        def __init__(self, key_size: _ty.Literal[192], pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "TripleDES", None) is not None:
    class _TripleDES_KEY(_SymmetricKeyBase, _TripleDES_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.TripleDES
        _FIXED_KEY_BYTES = 24

        def __init__(self, key_size: _ty.Literal[192], pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
else:
    class _TripleDES_KEY(_TripleDES_KEYTYPE): __concrete__ = False


if getattr(algorithms, "Blowfish", None) is not None:
    class _Blowfish_KEY(_SymmetricKeyBase, _Blowfish_KEYTYPE):
        _ALG_CLS = algorithms.Blowfish

        def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "Blowfish", None) is not None:
    class _Blowfish_KEY(_SymmetricKeyBase, _Blowfish_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.Blowfish

        def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
else:
    class _Blowfish_KEY(_Blowfish_KEYTYPE): __concrete__ = False


if getattr(algorithms, "CAST5", None) is not None:
    class _CAST5_KEY(_SymmetricKeyBase, _CAST5_KEYTYPE):
        _ALG_CLS = algorithms.CAST5

        def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "CAST5", None) is not None:
    class _CAST5_KEY(_SymmetricKeyBase, _CAST5_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.CAST5

        def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
else:
    class _CAST5_KEY(_CAST5_KEYTYPE): __concrete__ = False


if getattr(algorithms, "ARC4", None) is not None:
    class _ARC4_KEY(_SymmetricKeyBase, _ARC4_KEYTYPE):
        _ALG_CLS = algorithms.ARC4
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(128, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "ARC4", None) is not None:
    class _ARC4_KEY(_SymmetricKeyBase, _ARC4_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.ARC4
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(128, pwd)
else:
    class _ARC4_KEY(_ARC4_KEYTYPE): __concrete__ = False


if getattr(algorithms, "IDEA", None) is not None:
    class _IDEA_KEY(_SymmetricKeyBase, _IDEA_KEYTYPE):
        _ALG_CLS = algorithms.IDEA
        _FIXED_KEY_BYTES = 16
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(128, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "IDEA", None) is not None:
    class _IDEA_KEY(_SymmetricKeyBase, _IDEA_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.IDEA
        _FIXED_KEY_BYTES = 16
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(128, pwd)
else:
    class _IDEA_KEY(_IDEA_KEYTYPE): __concrete__ = False


if getattr(algorithms, "SEED", None) is not None:
    class _SEED_KEY(_SymmetricKeyBase, _SEED_KEYTYPE):
        _ALG_CLS = algorithms.SEED
        _FIXED_KEY_BYTES = 16
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(128, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "SEED", None) is not None:
    class _SEED_KEY(_SymmetricKeyBase, _SEED_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.SEED
        _FIXED_KEY_BYTES = 16
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(128, pwd)
else:
    class _SEED_KEY(_SEED_KEYTYPE): __concrete__ = False


if getattr(algorithms, "DES", None) is not None:
    class _DES_KEY(_SymmetricKeyBase, _DES_KEYTYPE):
        _ALG_CLS = algorithms.DES
        _FIXED_KEY_BYTES = 8
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(64, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "DES", None) is not None:
    class _DES_KEY(_SymmetricKeyBase, _DES_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.DES
        _FIXED_KEY_BYTES = 8
        _PWD_ONLY_INIT = True

        def __init__(self, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(64, pwd)
else:
    class _DES_KEY(_DES_KEYTYPE): __concrete__ = False


if getattr(algorithms, "ARC2", None) is not None:
    class _ARC2_KEY(_SymmetricKeyBase, _ARC2_KEYTYPE):
        _ALG_CLS = algorithms.ARC2

        def __init__(self, key_size: int = 128, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
elif decrepit_algorithms is not None and getattr(decrepit_algorithms, "ARC2", None) is not None:
    class _ARC2_KEY(_SymmetricKeyBase, _ARC2_KEYTYPE):
        _ALG_CLS = decrepit_algorithms.ARC2

        def __init__(self, key_size: int = 128, pwd: _ty.Optional[bytes | str] = None) -> None:
            super().__init__(key_size, pwd)
else:
    class _ARC2_KEY(_ARC2_KEYTYPE): __concrete__ = False


def _asym_encoding(encoding: ASymKeyEncoding) -> serialization.Encoding:
    if encoding == ASymKeyEncoding.PEM:
        return serialization.Encoding.PEM
    if encoding == ASymKeyEncoding.DER:
        return serialization.Encoding.DER
    if encoding == ASymKeyEncoding.RAW:
        return serialization.Encoding.Raw
    raise _NotSupportedError(f"Unsupported encoding: {encoding}")


def _private_format(format_: ASymKeyFormat, is_rsa: bool, is_ec: bool) -> serialization.PrivateFormat:
    if format_ == ASymKeyFormat.PKCS8:
        return serialization.PrivateFormat.PKCS8
    if format_ == ASymKeyFormat.PKCS1:
        if not is_rsa:
            raise _NotSupportedError("PKCS1 private encoding is RSA-only")
        return serialization.PrivateFormat.TraditionalOpenSSL
    if format_ == ASymKeyFormat.SEC1:
        if not is_ec:
            raise _NotSupportedError("SEC1 private encoding is EC-only")
        return serialization.PrivateFormat.TraditionalOpenSSL
    raise _NotSupportedError(f"Unsupported private key format: {format_}")


def _public_format(format_: ASymKeyFormat, encoding: ASymKeyEncoding, is_rsa: bool, is_ec: bool) -> serialization.PublicFormat:
    if format_ == ASymKeyFormat.OPENSSH:
        return serialization.PublicFormat.OpenSSH
    if format_ == ASymKeyFormat.PKCS1:
        if not is_rsa:
            raise _NotSupportedError("PKCS1 public encoding is RSA-only")
        return serialization.PublicFormat.PKCS1
    if format_ == ASymKeyFormat.SEC1:
        if not is_ec:
            raise _NotSupportedError("SEC1 public encoding is EC-only")
        if encoding == ASymKeyEncoding.RAW:
            return serialization.PublicFormat.CompressedPoint
        return serialization.PublicFormat.SubjectPublicKeyInfo
    return serialization.PublicFormat.SubjectPublicKeyInfo


def _load_private_key(data: bytes, encoding: ASymKeyEncoding, password: bytes | None):
    if encoding == ASymKeyEncoding.PEM:
        return serialization.load_pem_private_key(data, password=password, backend=default_backend())
    if encoding == ASymKeyEncoding.DER:
        return serialization.load_der_private_key(data, password=password, backend=default_backend())
    raise _NotSupportedError("RAW private key decode is unsupported for this key type")


def _load_public_key(data: bytes, encoding: ASymKeyEncoding):
    if encoding == ASymKeyEncoding.PEM:
        return serialization.load_pem_public_key(data, backend=default_backend())
    if encoding == ASymKeyEncoding.DER:
        return serialization.load_der_public_key(data, backend=default_backend())
    raise _NotSupportedError("RAW public key decode is unsupported for this key type")


def _rsa_padding(padding: ASymPadding):
    if padding == ASymPadding.OAEP:
        return asym_padding.OAEP(
            mgf=asym_padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        )
    if padding == ASymPadding.PKCShash1v15:
        return asym_padding.PKCS1v15()
    raise _NotSupportedError(f"Unsupported RSA padding: {padding}")


def _rsa_sign_padding(padding: ASymPadding):
    if padding == ASymPadding.PSS:
        return asym_padding.PSS(
            mgf=asym_padding.MGF1(hashes.SHA256()),
            salt_length=asym_padding.PSS.MAX_LENGTH,
        )
    if padding == ASymPadding.PKCShash1v15:
        return asym_padding.PKCS1v15()
    raise _NotSupportedError(f"Unsupported RSA signing padding: {padding}")


_CURVE_MAP: dict[ECCCurve, _ty.Type[ec.EllipticCurve]] = {
    ECCCurve.SECP192R1: ec.SECP192R1,
    ECCCurve.SECP224R1: ec.SECP224R1,
    ECCCurve.SECP256K1: ec.SECP256K1,
    ECCCurve.SECP256R1: ec.SECP256R1,
    ECCCurve.SECP384R1: ec.SECP384R1,
    ECCCurve.SECP521R1: ec.SECP521R1,
    ECCCurve.SECT163K1: ec.SECT163K1,
    ECCCurve.SECT163R2: ec.SECT163R2,
    ECCCurve.SECT233K1: ec.SECT233K1,
    ECCCurve.SECT233R1: ec.SECT233R1,
    ECCCurve.SECT283K1: ec.SECT283K1,
    ECCCurve.SECT283R1: ec.SECT283R1,
    ECCCurve.SECT409K1: ec.SECT409K1,
    ECCCurve.SECT409R1: ec.SECT409R1,
    ECCCurve.SECT571K1: ec.SECT571K1,
    ECCCurve.SECT571R1: ec.SECT571R1,
}


class _RSA_KEYPAIR(_RSA_KEYPAIRTYPE):
    backend = Backend.cryptography
    __concrete__ = True

    def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str] = None) -> None:
        if isinstance(pwd, str):
            pwd = pwd.encode("utf-8")
        self._private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size, backend=default_backend())
        self._public_key = self._private_key.public_key()

    @classmethod
    def decode_private_key(
        cls,
        data: bytes,
        format_: ASymKeyFormat,
        encoding: ASymKeyEncoding,
        password: bytes | None = None,
    ) -> _ty.Self:
        key = _load_private_key(data, encoding, password)
        if not isinstance(key, rsa.RSAPrivateKey):
            raise ValueError("Provided key is not an RSA private key")
        obj = cls(key.key_size)
        obj._private_key = key
        obj._public_key = key.public_key()
        return obj

    @classmethod
    def decode_public_key(
        cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding
    ) -> _ty.Self:
        key = _load_public_key(data, encoding)
        if not isinstance(key, rsa.RSAPublicKey):
            raise ValueError("Provided key is not an RSA public key")
        obj = cls(2048)
        obj._private_key = None
        obj._public_key = key
        return obj

    def encode_private_key(
        self,
        format_: ASymKeyFormat,
        encoding: ASymKeyEncoding,
        password: bytes | None = None,
    ) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        enc = _asym_encoding(encoding)
        private_format = _private_format(format_, is_rsa=True, is_ec=False)
        protection = (
            serialization.NoEncryption()
            if password is None
            else serialization.BestAvailableEncryption(password)
        )
        return self._private_key.private_bytes(enc, private_format, protection)

    def encode_public_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> bytes:
        if self._public_key is None:
            raise ValueError("No public key present")
        enc = _asym_encoding(encoding)
        public_format = _public_format(format_, encoding, is_rsa=True, is_ec=False)
        return self._public_key.public_bytes(enc, public_format)

    def encrypt(self, plain_bytes: bytes, padding: ASymPadding) -> bytes:
        if self._public_key is None:
            raise ValueError("No public key present")
        return self._public_key.encrypt(plain_bytes, _rsa_padding(padding))

    def decrypt(self, cipher_bytes: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        return self._private_key.decrypt(cipher_bytes, _rsa_padding(padding))

    def sign(self, data: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        return self._private_key.sign(data, _rsa_sign_padding(padding), hashes.SHA256())

    def sign_verify(self, data: bytes, signature: bytes, padding: ASymPadding) -> bool:
        if self._public_key is None:
            raise ValueError("No public key present")
        try:
            self._public_key.verify(signature, data, _rsa_sign_padding(padding), hashes.SHA256())
            return True
        except InvalidSignature:
            return False

    def __repr__(self) -> str:
        return f"<RSA private={self._private_key is not None}>"


class _DSA_KEYPAIR(_DSA_KEYPAIRTYPE):
    backend = Backend.cryptography
    __concrete__ = True

    def __init__(self, key_size: int, pwd: _ty.Optional[bytes | str] = None) -> None:
        self._private_key = dsa.generate_private_key(key_size=key_size, backend=default_backend())
        self._public_key = self._private_key.public_key()

    @classmethod
    def decode_private_key(
        cls,
        data: bytes,
        format_: ASymKeyFormat,
        encoding: ASymKeyEncoding,
        password: bytes | None = None,
    ) -> _ty.Self:
        key = _load_private_key(data, encoding, password)
        if not isinstance(key, dsa.DSAPrivateKey):
            raise ValueError("Provided key is not a DSA private key")
        obj = cls(key.key_size)
        obj._private_key = key
        obj._public_key = key.public_key()
        return obj

    @classmethod
    def decode_public_key(
        cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding
    ) -> _ty.Self:
        key = _load_public_key(data, encoding)
        if not isinstance(key, dsa.DSAPublicKey):
            raise ValueError("Provided key is not a DSA public key")
        obj = cls(2048)
        obj._private_key = None
        obj._public_key = key
        return obj

    def encode_private_key(
        self,
        format_: ASymKeyFormat,
        encoding: ASymKeyEncoding,
        password: bytes | None = None,
    ) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        if format_ == ASymKeyFormat.PKCS1:
            raise _NotSupportedError("PKCS1 is RSA-only")
        enc = _asym_encoding(encoding)
        protection = (
            serialization.NoEncryption()
            if password is None
            else serialization.BestAvailableEncryption(password)
        )
        return self._private_key.private_bytes(
            enc,
            serialization.PrivateFormat.PKCS8,
            protection,
        )

    def encode_public_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> bytes:
        if self._public_key is None:
            raise ValueError("No public key present")
        if format_ == ASymKeyFormat.PKCS1:
            raise _NotSupportedError("PKCS1 is RSA-only")
        enc = _asym_encoding(encoding)
        public_format = _public_format(format_, encoding, is_rsa=False, is_ec=False)
        return self._public_key.public_bytes(enc, public_format)

    def sign(self, data: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        return self._private_key.sign(data, hashes.SHA256())

    def sign_verify(self, data: bytes, signature: bytes, padding: ASymPadding) -> bool:
        if self._public_key is None:
            raise ValueError("No public key present")
        try:
            self._public_key.verify(signature, data, hashes.SHA256())
            return True
        except InvalidSignature:
            return False

    def __repr__(self) -> str:
        return f"<DSA private={self._private_key is not None}>"


class _ECC_KEYPAIR(_ECC_KEYPAIRTYPE):
    backend = Backend.cryptography
    __concrete__ = True

    def __init__(
        self,
        ecc_type: ECCType = ECCType.ECDSA,
        ecc_curve: ECCCurve | None = ECCCurve.SECP256R1,
        pwd: _ty.Optional[bytes | str] = None,
    ) -> None:
        self._ecc_type = ecc_type
        self._ecc_curve = ecc_curve
        self._private_key: _ty.Any
        self._public_key: _ty.Any

        if ecc_type == ECCType.ECDSA:
            curve_cls = _CURVE_MAP.get(ecc_curve or ECCCurve.SECP256R1)
            if curve_cls is None:
                raise ValueError(f"Unsupported ECC curve: {ecc_curve}")
            self._private_key = ec.generate_private_key(curve_cls(), backend=default_backend())
        elif ecc_type == ECCType.Ed25519:
            self._private_key = ed25519.Ed25519PrivateKey.generate()
        elif ecc_type == ECCType.Ed448:
            self._private_key = ed448.Ed448PrivateKey.generate()
        elif ecc_type == ECCType.X25519:
            self._private_key = x25519.X25519PrivateKey.generate()
        elif ecc_type == ECCType.X448:
            self._private_key = x448.X448PrivateKey.generate()
        else:
            raise ValueError(f"Unsupported ECC type: {ecc_type}")

        self._public_key = self._private_key.public_key()

    @classmethod
    def decode_private_key(
        cls,
        data: bytes,
        format_: ASymKeyFormat,
        encoding: ASymKeyEncoding,
        password: bytes | None = None,
    ) -> _ty.Self:
        key = _load_private_key(data, encoding, password)
        obj = cls(ECCType.ECDSA, ECCCurve.SECP256R1)
        if isinstance(key, ec.EllipticCurvePrivateKey):
            curve_name = key.curve.name.lower()
            curve = next((c for c, cc in _CURVE_MAP.items() if cc().name.lower() == curve_name), ECCCurve.SECP256R1)
            obj._ecc_type = ECCType.ECDSA
            obj._ecc_curve = curve
        elif isinstance(key, ed25519.Ed25519PrivateKey):
            obj._ecc_type = ECCType.Ed25519
            obj._ecc_curve = None
        elif isinstance(key, ed448.Ed448PrivateKey):
            obj._ecc_type = ECCType.Ed448
            obj._ecc_curve = None
        elif isinstance(key, x25519.X25519PrivateKey):
            obj._ecc_type = ECCType.X25519
            obj._ecc_curve = None
        elif isinstance(key, x448.X448PrivateKey):
            obj._ecc_type = ECCType.X448
            obj._ecc_curve = None
        else:
            raise ValueError("Provided key is not a supported ECC private key")

        obj._private_key = key
        obj._public_key = key.public_key()
        return obj

    @classmethod
    def decode_public_key(
        cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding
    ) -> _ty.Self:
        if encoding == ASymKeyEncoding.RAW:
            raise _NotSupportedError("RAW public decode requires type context for ECC")
        key = _load_public_key(data, encoding)
        obj = cls(ECCType.ECDSA, ECCCurve.SECP256R1)
        obj._private_key = None
        obj._public_key = key

        if isinstance(key, ec.EllipticCurvePublicKey):
            obj._ecc_type = ECCType.ECDSA
            curve_name = key.curve.name.lower()
            obj._ecc_curve = next((c for c, cc in _CURVE_MAP.items() if cc().name.lower() == curve_name), ECCCurve.SECP256R1)
        elif isinstance(key, ed25519.Ed25519PublicKey):
            obj._ecc_type = ECCType.Ed25519
            obj._ecc_curve = None
        elif isinstance(key, ed448.Ed448PublicKey):
            obj._ecc_type = ECCType.Ed448
            obj._ecc_curve = None
        elif isinstance(key, x25519.X25519PublicKey):
            obj._ecc_type = ECCType.X25519
            obj._ecc_curve = None
        elif isinstance(key, x448.X448PublicKey):
            obj._ecc_type = ECCType.X448
            obj._ecc_curve = None
        else:
            raise ValueError("Provided key is not a supported ECC public key")

        return obj

    def encode_private_key(
        self,
        format_: ASymKeyFormat,
        encoding: ASymKeyEncoding,
        password: bytes | None = None,
    ) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")

        if encoding == ASymKeyEncoding.RAW:
            if hasattr(self._private_key, "private_bytes_raw"):
                return self._private_key.private_bytes_raw()
            raise _NotSupportedError("RAW private encoding unsupported for this ECC type")

        enc = _asym_encoding(encoding)
        is_ec = self._ecc_type == ECCType.ECDSA
        priv_format = _private_format(format_, is_rsa=False, is_ec=is_ec)
        protection = (
            serialization.NoEncryption()
            if password is None
            else serialization.BestAvailableEncryption(password)
        )
        return self._private_key.private_bytes(enc, priv_format, protection)

    def encode_public_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> bytes:
        if self._public_key is None:
            raise ValueError("No public key present")

        if encoding == ASymKeyEncoding.RAW:
            if hasattr(self._public_key, "public_bytes_raw"):
                return self._public_key.public_bytes_raw()
            if self._ecc_type == ECCType.ECDSA:
                return self._public_key.public_bytes(
                    serialization.Encoding.X962,
                    serialization.PublicFormat.CompressedPoint,
                )
            raise _NotSupportedError("RAW public encoding unsupported for this ECC type")

        enc = _asym_encoding(encoding)
        pub_format = _public_format(
            format_, encoding, is_rsa=False, is_ec=(self._ecc_type == ECCType.ECDSA)
        )
        return self._public_key.public_bytes(enc, pub_format)

    def sign(self, data: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")

        if self._ecc_type == ECCType.ECDSA:
            return self._private_key.sign(data, ec.ECDSA(hashes.SHA256()))
        if self._ecc_type in (ECCType.Ed25519, ECCType.Ed448):
            return self._private_key.sign(data)
        raise _NotSupportedError(f"{self._ecc_type} does not support signing")

    def sign_verify(self, data: bytes, signature: bytes, padding: ASymPadding) -> bool:
        if self._public_key is None:
            raise ValueError("No public key present")

        try:
            if self._ecc_type == ECCType.ECDSA:
                self._public_key.verify(signature, data, ec.ECDSA(hashes.SHA256()))
            elif self._ecc_type in (ECCType.Ed25519, ECCType.Ed448):
                self._public_key.verify(signature, data)
            else:
                raise _NotSupportedError(f"{self._ecc_type} does not support signature verification")
            return True
        except InvalidSignature:
            return False

    def key_exchange(self, peer_public_key: _ty.Self) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")

        if self._ecc_type == ECCType.ECDSA:
            if not isinstance(self._private_key, ec.EllipticCurvePrivateKey):
                raise ValueError("Invalid private key for ECDSA key exchange")
            if not isinstance(peer_public_key._public_key, ec.EllipticCurvePublicKey):
                raise ValueError("Peer key must be an EC public key")
            return self._private_key.exchange(ec.ECDH(), peer_public_key._public_key)

        if self._ecc_type == ECCType.X25519:
            if not isinstance(peer_public_key._public_key, x25519.X25519PublicKey):
                raise ValueError("Peer key must be an X25519 public key")
            return self._private_key.exchange(peer_public_key._public_key)

        if self._ecc_type == ECCType.X448:
            if not isinstance(peer_public_key._public_key, x448.X448PublicKey):
                raise ValueError("Peer key must be an X448 public key")
            return self._private_key.exchange(peer_public_key._public_key)

        raise _NotSupportedError(f"{self._ecc_type} does not support key exchange")

    def __repr__(self) -> str:
        return f"<ECC type={self._ecc_type}>"



from __future__ import annotations

import pytest

pytest.importorskip("Cryptodome")

from ..crypto import Backend, set_backend
from ..crypto.algos import Asym, HashAlgorithm, KeyDerivationFunction, Sym
from ..crypto.exceptions import NotSupportedError


@pytest.fixture()
def pycrypto_backend() -> None:
    set_backend([Backend.pycryptodomex_alpha, Backend.std_lib])


def test_backend_wires_core(pycrypto_backend: None) -> None:
    assert Sym.Cipher.AES.key.__concrete__ is True
    assert Sym.Cipher.ChaCha20.key.__concrete__ is True
    assert Asym.Cipher.RSA.keypair.__concrete__ is True


def test_hash_roundtrip(pycrypto_backend: None) -> None:
    for hasher in (HashAlgorithm.SHA1, HashAlgorithm.SHA2.SHA256, HashAlgorithm.BLAKE2.BLAKE2s):
        h = hasher.hash(b"abc")
        assert hasher.verify(b"abc", h)


def test_kdf_roundtrip(pycrypto_backend: None) -> None:
    salt = b"0123456789abcdef"
    assert len(KeyDerivationFunction.PBKDF2HMAC.derive(b"pw", salt=salt, length=16)) == 16
    assert len(KeyDerivationFunction.Scrypt.derive(b"pw", salt=salt, length=16, n=2**10)) == 16
    assert len(KeyDerivationFunction.HKDF.derive(b"pw", salt=salt, length=16)) == 16


def test_aes_gcm_and_cbc(pycrypto_backend: None) -> None:
    key = Sym.Cipher.AES.key.new(128)
    pt = b"hello"

    gcm = key.encrypt(pt, Sym.Padding.PKCS7, Sym.Operation.GCM)
    assert key.decrypt(gcm, Sym.Padding.PKCS7, Sym.Operation.GCM) == pt

    cbc = key.encrypt(pt, Sym.Padding.PKCS7, Sym.Operation.CBC)
    assert key.decrypt(cbc, Sym.Padding.PKCS7, Sym.Operation.CBC) == pt


def test_chacha20_roundtrip(pycrypto_backend: None) -> None:
    key = Sym.Cipher.ChaCha20.key.new()
    pt = b"hello stream"
    ct = key.encrypt(pt, Sym.Padding.PKCS7, Sym.Operation.CTR)
    assert key.decrypt(ct, Sym.Padding.PKCS7, Sym.Operation.CTR) == pt



def test_mac_support(pycrypto_backend: None) -> None:
    key = Sym.Cipher.AES.key.new(128)
    data = b"auth"
    m = key.generate_mac(data, Sym.MessageAuthenticationCode.HMAC)
    assert key.verify_mac(data, m, Sym.MessageAuthenticationCode.HMAC)


def test_rsa_sign_encrypt(pycrypto_backend: None) -> None:
    key = Asym.Cipher.RSA.keypair.new(2048)
    msg = b"hello"

    sig = key.sign(msg, Asym.Padding.PSS)
    assert key.sign_verify(msg, sig, Asym.Padding.PSS)

    ct = key.encrypt(msg, Asym.Padding.OAEP)
    assert key.decrypt(ct, Asym.Padding.OAEP) == msg


def test_dsa_sign(pycrypto_backend: None) -> None:
    key = Asym.Cipher.DSA.keypair.new(2048)
    msg = b"hello"
    sig = key.sign(msg, Asym.Padding.PSS)
    assert key.sign_verify(msg, sig, Asym.Padding.PSS)


def test_ecc_sign_and_key_exchange(pycrypto_backend: None) -> None:
    a = Asym.Cipher.ECC.keypair.ecdsa_key(Asym.Cipher.ECC.Curve.SECP256R1)
    b = Asym.Cipher.ECC.keypair.ecdsa_key(Asym.Cipher.ECC.Curve.SECP256R1)

    msg = b"hello"
    sig = a.sign(msg, Asym.Padding.PSS)
    assert a.sign_verify(msg, sig, Asym.Padding.PSS)

    try:
        sa = a.key_exchange(b)
        sb = b.key_exchange(a)
        assert sa == sb
    except NotSupportedError:
        pytest.skip("ECC key exchange not available in this pycryptodomex build")


def test_fallback_hash(pycrypto_backend: None) -> None:
    h = HashAlgorithm.SHA2.SHA256.hash(b"hello")
    assert HashAlgorithm.SHA2.SHA256.verify(b"hello", h)



from quantcrypt.kem import MLKEM_512, MLKEM_768, MLKEM_1024
from quantcrypt.dss import MLDSA_44, MLDSA_65, MLDSA_87
from quantcrypt import errors


from .._definitions import _KYBER_KEYPAIRTYPE, _DILITHIUM_KEYPAIRTYPE#, SymKeyEncoding
from ..exceptions import NotSupportedError as _NotSupportedError
from ..algos._asym import (
    KeyFormat as ASymKeyFormat,
    KeyEncoding as ASymKeyEncoding,
    Padding as ASymPadding,
)

import typing_extensions as _te
import typing as _ty


_KYBER_VARIANTS = {
    "kyber512": MLKEM_512,
    "kyber768": MLKEM_768,
    "kyber1024": MLKEM_1024,
}


class _KYBER_KEYPAIR(_KYBER_KEYPAIRTYPE):
    def __init__(self, mode: _ty.Literal["kyber512", "kyber768", "kyber1024"]) -> None:
        try:
            impl_cls = _KYBER_VARIANTS[mode]
        except KeyError:
            raise ValueError(f"Unsupported Kyber mode: {mode}")
        self._mode = mode
        self._impl = impl_cls()
        self._public_key: bytes | None
        self._private_key: bytes | None
        self._public_key, self._private_key = self._impl.keygen()

    @classmethod
    def decode(
        cls,
        mode: _ty.Literal["kyber512", "kyber768", "kyber1024"],
        *,
        public_key: bytes | None = None,
        private_key: bytes | None = None,
    ) -> _te.Self:
        """, encoding: SymKeyEncoding"""
        obj = cls(mode)
        obj._public_key = public_key
        obj._private_key = private_key
        return obj

    def encode_private_key(self) -> bytes | None:
        return self._private_key

    def encode_public_key(self) -> bytes | None:
        return self._public_key

    def encapsulate(self) -> tuple[bytes, bytes]:
        if self._public_key is None:
            raise ValueError("Cannot encaps without the public key")
        return self._impl.encaps(self._public_key)

    def decapsulate(self, ciphertext: bytes) -> bytes:
        if self._private_key is None:
            raise ValueError("Cannot encaps without the private key")
        return self._impl.decaps(self._private_key, ciphertext)

    def __repr__(self) -> str:
        return f"<KYBER mode={self._mode}>"


_DILITHIUM_VARIANTS = {
    "dilithium2": MLDSA_44,
    "dilithium3": MLDSA_65,
    "dilithium5": MLDSA_87,
}


class _DILITHIUM_KEYPAIR(_DILITHIUM_KEYPAIRTYPE):
    def __init__(
        self, mode: _ty.Literal["dilithium2", "dilithium3", "dilithium5"]
    ) -> None:
        try:
            impl_cls = _DILITHIUM_VARIANTS[mode]
        except KeyError:
            raise ValueError(f"Unsupported Dilithium mode: {mode}")
        self._mode = mode
        self._impl = impl_cls()
        self._public_key: bytes | None
        self._private_key: bytes | None
        self._public_key, self._private_key = self._impl.keygen()

    @classmethod
    def decode(
        cls,
        mode: _ty.Literal["dilithium2", "dilithium3", "dilithium5"],
        *,
        public_key: bytes | None = None,
        private_key: bytes | None = None,
    ) -> _te.Self:
        """, encoding: SymKeyEncoding"""
        if public_key is None or private_key is None:
            raise ValueError("Please pass both public and private key for now as this backend does not support public "
                             "key regeneration as of now.")
        obj = cls(mode)
        obj._public_key = public_key
        obj._private_key = private_key
        return obj

    def encode_private_key(self) -> bytes | None:
        return self._private_key

    def encode_public_key(self) -> bytes | None:
        return self._public_key

    def sign(self, data: bytes) -> bytes:
        if self._private_key is None:
            raise ValueError("Cannot sign without the private key")
        return self._impl.sign(self._private_key, data)

    def sign_verify(self, data: bytes, signature: bytes) -> bool:
        if self._public_key is None:
            raise ValueError("Cannot verify sign without the public key")
        return self._impl.verify(self._public_key, data, signature)

    def __repr__(self) -> str:
        return f"<DILITHIUM mode={self._mode}>"



from __future__ import annotations

import base64
import hashlib
import os
import typing as _ty

from Cryptodome.Cipher import AES, ARC2, ARC4, Blowfish, CAST, ChaCha20, DES, DES3, PKCS1_OAEP, PKCS1_v1_5, Salsa20
from Cryptodome.Hash import CMAC, HMAC, SHA256
from Cryptodome.Protocol import DH
from Cryptodome.PublicKey import DSA, ECC, RSA
from Cryptodome.Signature import DSS, eddsa, pkcs1_15, pss
from Cryptodome.Util.Padding import pad, unpad

try:
    from Cryptodome.Cipher import Camellia
except Exception:  # pragma: no cover
    Camellia = None

try:
    from Cryptodome.Cipher import IDEA
except Exception:  # pragma: no cover
    IDEA = None

try:
    from Cryptodome.Cipher import SEED
except Exception:  # pragma: no cover
    SEED = None

try:
    from Cryptodome.Cipher import SM4
except Exception:  # pragma: no cover
    SM4 = None

from .._definitions import (
    Backend,
    ECCCurve,
    ECCType,
    _AES_KEYTYPE,
    _ARC2_KEYTYPE,
    _ARC4_KEYTYPE,
    _Blowfish_KEYTYPE,
    _CAST5_KEYTYPE,
    _Camellia_KEYTYPE,
    _ChaCha20_KEYTYPE,
    _DES_KEYTYPE,
    _DSA_KEYPAIRTYPE,
    _ECC_KEYPAIRTYPE,
    _IDEA_KEYTYPE,
    _RSA_KEYPAIRTYPE,
    _Salsa20_KEYTYPE,
    _SEED_KEYTYPE,
    _SM4_KEYTYPE,
    _TripleDES_KEYTYPE,
)
from ..algos._asym import KeyEncoding as ASymKeyEncoding, KeyFormat as ASymKeyFormat, Padding as ASymPadding
from ..algos._sym import (
    KeyEncoding as SymKeyEncoding,
    MessageAuthenticationCode as MAC,
    Operation as SymOperation,
    Padding as SymPadding,
)
from ..exceptions import NotSupportedError as _NotSupportedError

_BLOCK_CIPHERS: dict[type, tuple[_ty.Any, tuple[int, ...] | None]] = {
    _AES_KEYTYPE: (AES, (16, 24, 32)),
    _TripleDES_KEYTYPE: (DES3, (24,)),
    _Blowfish_KEYTYPE: (Blowfish, None),
    _CAST5_KEYTYPE: (CAST, None),
    _DES_KEYTYPE: (DES, (8,)),
    _ARC2_KEYTYPE: (ARC2, None),
}
if Camellia is not None:
    _BLOCK_CIPHERS[_Camellia_KEYTYPE] = (Camellia, None)
if IDEA is not None:
    _BLOCK_CIPHERS[_IDEA_KEYTYPE] = (IDEA, (16,))
if SEED is not None:
    _BLOCK_CIPHERS[_SEED_KEYTYPE] = (SEED, (16,))
if SM4 is not None:
    _BLOCK_CIPHERS[_SM4_KEYTYPE] = (SM4, (16,))

_STREAM_CIPHERS: dict[type, tuple[_ty.Any, int | None]] = {
    _ARC4_KEYTYPE: (ARC4, None),
    _ChaCha20_KEYTYPE: (ChaCha20, 32),
    _Salsa20_KEYTYPE: (Salsa20, None),
}


def _derive_key_from_password(password: bytes | str, length: int, salt: bytes) -> bytes:
    if isinstance(password, str):
        password = password.encode("utf-8")
    return hashlib.pbkdf2_hmac("sha256", password, salt, 100_000, dklen=length)


def _encode_bytes(data: bytes, encoding: SymKeyEncoding) -> bytes:
    if encoding == SymKeyEncoding.RAW:
        return data
    if encoding == SymKeyEncoding.BASE64:
        return base64.b64encode(data)
    if encoding == SymKeyEncoding.HEX:
        return data.hex().encode("ascii")
    if encoding == SymKeyEncoding.BASE32:
        return base64.b32encode(data)
    raise _NotSupportedError(f"Unsupported key encoding: {encoding}")


def _decode_bytes(data: bytes, encoding: SymKeyEncoding) -> bytes:
    if encoding == SymKeyEncoding.RAW:
        return data
    if encoding == SymKeyEncoding.BASE64:
        return base64.b64decode(data)
    if encoding == SymKeyEncoding.HEX:
        return bytes.fromhex(data.decode("ascii"))
    if encoding == SymKeyEncoding.BASE32:
        return base64.b32decode(data)
    raise _NotSupportedError(f"Unsupported key encoding: {encoding}")


def _pack(parts: dict[str, bytes]) -> bytes:
    iv = parts.get("iv", b"")
    tag = parts.get("tag", b"")
    ct = parts["ciphertext"]
    return len(iv).to_bytes(2, "big") + len(tag).to_bytes(2, "big") + iv + tag + ct


def _unpack(payload: bytes) -> dict[str, bytes]:
    if len(payload) < 4:
        raise ValueError("cipher payload too short")
    iv_len = int.from_bytes(payload[:2], "big")
    tag_len = int.from_bytes(payload[2:4], "big")
    if len(payload) < 4 + iv_len + tag_len:
        raise ValueError("cipher payload malformed")
    base = 4
    iv = payload[base:base + iv_len]
    base += iv_len
    tag = payload[base:base + tag_len]
    ct = payload[base + tag_len:]
    out = {"ciphertext": ct}
    if iv_len:
        out["iv"] = iv
    if tag_len:
        out["tag"] = tag
    return out


class _SymBase:
    backend = Backend.pycryptodomex
    __concrete__ = True

    _ciphermod: _ty.Any
    _stream: bool = False
    _fixed_len: int | None = None
    _allow_gcm: bool = False

    def __init__(self, key_size: int, pwd: bytes | str | None) -> None:
        key_len = self._fixed_len if self._fixed_len is not None else key_size // 8
        self.key_size = key_len * 8

        if pwd is None:
            self._key = os.urandom(key_len)
        else:
            pwd_b = pwd.encode("utf-8") if isinstance(pwd, str) else pwd
            if len(pwd_b) == key_len:
                self._key = pwd_b
            else:
                self._key = _derive_key_from_password(pwd_b, key_len, hashlib.sha256(pwd_b).digest()[:16])

    @classmethod
    def decode(cls, key: bytes, encoding: SymKeyEncoding) -> _ty.Self:
        raw = _decode_bytes(key, encoding)
        size_bits = len(raw) * 8
        try:
            obj = cls(raw)  # for pwd-only constructors
        except TypeError:
            obj = cls(size_bits, raw)
        return obj

    def encode(self, encoding: SymKeyEncoding) -> bytes:
        return _encode_bytes(self._key, encoding)

    def _block_size(self) -> int:
        if self._stream:
            return 1
        return self._ciphermod.block_size

    def _pad(self, data: bytes, padding: SymPadding) -> bytes:
        if padding == SymPadding.PKCS7:
            return pad(data, self._block_size(), style="pkcs7")
        if padding == SymPadding.ANSIX923:
            return pad(data, self._block_size(), style="x923")
        raise _NotSupportedError(f"Unsupported padding {padding}")

    def _unpad(self, data: bytes, padding: SymPadding) -> bytes:
        if padding == SymPadding.PKCS7:
            return unpad(data, self._block_size(), style="pkcs7")
        if padding == SymPadding.ANSIX923:
            return unpad(data, self._block_size(), style="x923")
        raise _NotSupportedError(f"Unsupported padding {padding}")

    def _cipher_for_encrypt(self, mode: SymOperation):
        if self._stream:
            if self._ciphermod is ARC4:
                return self._ciphermod.new(self._key), {}
            if self._ciphermod is ChaCha20:
                nonce = os.urandom(8)
                return self._ciphermod.new(key=self._key, nonce=nonce), {"iv": nonce}
            if self._ciphermod is Salsa20:
                nonce = os.urandom(8)
                return self._ciphermod.new(key=self._key, nonce=nonce), {"iv": nonce}

        if mode == SymOperation.ECB:
            return self._ciphermod.new(self._key, self._ciphermod.MODE_ECB), {}
        if mode == SymOperation.CBC:
            iv = os.urandom(self._ciphermod.block_size)
            return self._ciphermod.new(self._key, self._ciphermod.MODE_CBC, iv=iv), {"iv": iv}
        if mode == SymOperation.CFB:
            iv = os.urandom(self._ciphermod.block_size)
            return self._ciphermod.new(self._key, self._ciphermod.MODE_CFB, iv=iv, segment_size=self._ciphermod.block_size * 8), {"iv": iv}
        if mode == SymOperation.OFB:
            iv = os.urandom(self._ciphermod.block_size)
            return self._ciphermod.new(self._key, self._ciphermod.MODE_OFB, iv=iv), {"iv": iv}
        if mode == SymOperation.CTR:
            nonce = os.urandom(max(1, self._ciphermod.block_size // 2))
            return self._ciphermod.new(self._key, self._ciphermod.MODE_CTR, nonce=nonce), {"iv": nonce}
        if mode == SymOperation.GCM:
            if not self._allow_gcm:
                raise _NotSupportedError("GCM is only supported for AES")
            nonce = os.urandom(12)
            return self._ciphermod.new(self._key, self._ciphermod.MODE_GCM, nonce=nonce), {"iv": nonce}
        raise _NotSupportedError(f"Unsupported operation mode {mode}")

    def _cipher_for_decrypt(self, mode: SymOperation, parts: dict[str, bytes]):
        iv = parts.get("iv")
        tag = parts.get("tag")

        if self._stream:
            if self._ciphermod is ARC4:
                return self._ciphermod.new(self._key)
            if self._ciphermod is ChaCha20:
                if iv is None:
                    raise ValueError("ChaCha20 requires nonce")
                return self._ciphermod.new(key=self._key, nonce=iv)
            if self._ciphermod is Salsa20:
                if iv is None:
                    raise ValueError("Salsa20 requires nonce")
                return self._ciphermod.new(key=self._key, nonce=iv)

        if mode == SymOperation.ECB:
            return self._ciphermod.new(self._key, self._ciphermod.MODE_ECB)
        if mode == SymOperation.CBC:
            if iv is None:
                raise ValueError("CBC requires IV")
            return self._ciphermod.new(self._key, self._ciphermod.MODE_CBC, iv=iv)
        if mode == SymOperation.CFB:
            if iv is None:
                raise ValueError("CFB requires IV")
            return self._ciphermod.new(self._key, self._ciphermod.MODE_CFB, iv=iv, segment_size=self._ciphermod.block_size * 8)
        if mode == SymOperation.OFB:
            if iv is None:
                raise ValueError("OFB requires IV")
            return self._ciphermod.new(self._key, self._ciphermod.MODE_OFB, iv=iv)
        if mode == SymOperation.CTR:
            if iv is None:
                raise ValueError("CTR requires nonce")
            return self._ciphermod.new(self._key, self._ciphermod.MODE_CTR, nonce=iv)
        if mode == SymOperation.GCM:
            if not self._allow_gcm:
                raise _NotSupportedError("GCM is only supported for AES")
            if iv is None or tag is None:
                raise ValueError("GCM requires nonce and tag")
            return self._ciphermod.new(self._key, self._ciphermod.MODE_GCM, nonce=iv)
        raise _NotSupportedError(f"Unsupported operation mode {mode}")

    def encrypt(self, plain_bytes: bytes, padding: SymPadding, mode: SymOperation, /, *, auto_pack: bool = True) -> bytes | dict[str, bytes]:
        if self._stream and mode is not SymOperation.CTR:
            raise _NotSupportedError("Stream ciphers support only CTR-like mode in this API")

        cipher, meta = self._cipher_for_encrypt(mode)
        payload = plain_bytes if self._stream or mode not in (SymOperation.ECB, SymOperation.CBC) else self._pad(plain_bytes, padding)
        ciphertext = cipher.encrypt(payload)
        parts = {"ciphertext": ciphertext, **meta}
        if mode == SymOperation.GCM and hasattr(cipher, "digest"):
            parts["tag"] = cipher.digest()
        return _pack(parts) if auto_pack else parts

    def decrypt(self, cipher_bytes_or_dict: bytes | dict[str, bytes], padding: SymPadding, mode: SymOperation, /, *, auto_pack: bool = True) -> bytes:
        if self._stream and mode is not SymOperation.CTR:
            raise _NotSupportedError("Stream ciphers support only CTR-like mode in this API")

        parts = _unpack(cipher_bytes_or_dict) if auto_pack else _ty.cast(dict[str, bytes], cipher_bytes_or_dict)
        cipher = self._cipher_for_decrypt(mode, parts)
        plaintext = cipher.decrypt(parts["ciphertext"])

        if mode == SymOperation.GCM:
            cipher.verify(parts["tag"])
        if not self._stream and mode in (SymOperation.ECB, SymOperation.CBC):
            plaintext = self._unpad(plaintext, padding)
        return plaintext

    def generate_mac(self, data: bytes, auth_type: MAC) -> bytes:
        if auth_type == MAC.HMAC:
            h = HMAC.new(self._key, digestmod=SHA256)
            h.update(data)
            return h.digest()
        if auth_type == MAC.CMAC:
            if self._stream:
                raise _NotSupportedError("CMAC is only supported for block ciphers")
            c = CMAC.new(self._key, ciphermod=self._ciphermod)
            c.update(data)
            return c.digest()
        raise _NotSupportedError(f"Unsupported MAC type {auth_type}")

    def verify_mac(self, data: bytes, mac: bytes, auth_type: MAC) -> bool:
        return self.generate_mac(data, auth_type) == mac


class _AES_KEY(_SymBase, _AES_KEYTYPE):
    _ciphermod = AES
    _allow_gcm = True


class _ChaCha20_KEY(_SymBase, _ChaCha20_KEYTYPE):
    _ciphermod = ChaCha20
    _stream = True
    _fixed_len = 32

    def __init__(self, pwd: bytes | str | None = None) -> None:
        super().__init__(256, pwd)


class _TripleDES_KEY(_SymBase, _TripleDES_KEYTYPE):
    _ciphermod = DES3
    _fixed_len = 24

    def __init__(self, key_size: _ty.Literal[192], pwd: bytes | str | None = None) -> None:
        super().__init__(key_size, pwd)


class _Blowfish_KEY(_SymBase, _Blowfish_KEYTYPE):
    _ciphermod = Blowfish


class _CAST5_KEY(_SymBase, _CAST5_KEYTYPE):
    _ciphermod = CAST


class _ARC4_KEY(_SymBase, _ARC4_KEYTYPE):
    _ciphermod = ARC4
    _stream = True

    def __init__(self, pwd: bytes | str | None = None) -> None:
        super().__init__(128, pwd)


if Camellia is not None:
    class _Camellia_KEY(_SymBase, _Camellia_KEYTYPE):
        _ciphermod = Camellia
else:
    class _Camellia_KEY(_Camellia_KEYTYPE): __concrete__ = False


if IDEA is not None:
    class _IDEA_KEY(_SymBase, _IDEA_KEYTYPE):
        _ciphermod = IDEA
        _fixed_len = 16

        def __init__(self, pwd: bytes | str | None = None) -> None:
            super().__init__(128, pwd)
else:
    class _IDEA_KEY(_IDEA_KEYTYPE): __concrete__ = False


if SEED is not None:
    class _SEED_KEY(_SymBase, _SEED_KEYTYPE):
        _ciphermod = SEED
        _fixed_len = 16

        def __init__(self, pwd: bytes | str | None = None) -> None:
            super().__init__(128, pwd)
else:
    class _SEED_KEY(_SEED_KEYTYPE): __concrete__ = False


if SM4 is not None:
    class _SM4_KEY(_SymBase, _SM4_KEYTYPE):
        _ciphermod = SM4
        _fixed_len = 16

        def __init__(self, pwd: bytes | str | None = None) -> None:
            super().__init__(128, pwd)
else:
    class _SM4_KEY(_SM4_KEYTYPE): __concrete__ = False


class _DES_KEY(_SymBase, _DES_KEYTYPE):
    _ciphermod = DES
    _fixed_len = 8

    def __init__(self, pwd: bytes | str | None = None) -> None:
        super().__init__(64, pwd)


class _ARC2_KEY(_SymBase, _ARC2_KEYTYPE):
    _ciphermod = ARC2


class _Salsa20_KEY(_SymBase, _Salsa20_KEYTYPE):
    _ciphermod = Salsa20
    _stream = True


class _RSA_KEYPAIR(_RSA_KEYPAIRTYPE):
    backend = Backend.pycryptodomex
    __concrete__ = True

    def __init__(self, key_size: int, pwd: bytes | str | None = None) -> None:
        self._private_key = RSA.generate(key_size)
        self._public_key = self._private_key.publickey()

    @classmethod
    def decode_private_key(cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding, password: bytes | None = None) -> _ty.Self:
        key = RSA.import_key(data, passphrase=password)
        obj = cls(key.size_in_bits())
        obj._private_key = key
        obj._public_key = key.publickey()
        return obj

    @classmethod
    def decode_public_key(cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> _ty.Self:
        key = RSA.import_key(data)
        obj = cls(2048)
        obj._private_key = None
        obj._public_key = key
        return obj

    def encode_private_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding, password: bytes | None = None) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        fmt = "PEM" if encoding == ASymKeyEncoding.PEM else "DER"
        if format_ == ASymKeyFormat.PKCS8:
            if password is None:
                return self._private_key.export_key(format=fmt, pkcs=8)
            return self._private_key.export_key(format=fmt, pkcs=8, passphrase=password, protection="PBKDF2WithHMAC-SHA1AndAES256-CBC")
        if format_ == ASymKeyFormat.PKCS1:
            return self._private_key.export_key(format=fmt, pkcs=1, passphrase=password)
        raise _NotSupportedError(f"Unsupported RSA private format: {format_}")

    def encode_public_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> bytes:
        if self._public_key is None:
            raise ValueError("No public key present")
        if format_ == ASymKeyFormat.OPENSSH:
            return self._public_key.export_key(format="OpenSSH")
        fmt = "PEM" if encoding == ASymKeyEncoding.PEM else "DER"
        return self._public_key.export_key(format=fmt, pkcs=1 if format_ == ASymKeyFormat.PKCS1 else 8)

    def encrypt(self, plain_bytes: bytes, padding: ASymPadding) -> bytes:
        if padding == ASymPadding.OAEP:
            return PKCS1_OAEP.new(self._public_key).encrypt(plain_bytes)
        if padding == ASymPadding.PKCShash1v15:
            return PKCS1_v1_5.new(self._public_key).encrypt(plain_bytes)
        raise _NotSupportedError(f"Unsupported RSA encryption padding: {padding}")

    def decrypt(self, cipher_bytes: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        if padding == ASymPadding.OAEP:
            return PKCS1_OAEP.new(self._private_key).decrypt(cipher_bytes)
        if padding == ASymPadding.PKCShash1v15:
            sentinel = os.urandom(32)
            out = PKCS1_v1_5.new(self._private_key).decrypt(cipher_bytes, sentinel)
            if out == sentinel:
                raise ValueError("RSA PKCS1v15 decryption failed")
            return out
        raise _NotSupportedError(f"Unsupported RSA encryption padding: {padding}")

    def sign(self, data: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        h = SHA256.new(data)
        if padding == ASymPadding.PSS:
            return pss.new(self._private_key).sign(h)
        if padding == ASymPadding.PKCShash1v15:
            return pkcs1_15.new(self._private_key).sign(h)
        raise _NotSupportedError(f"Unsupported RSA signing padding: {padding}")

    def sign_verify(self, data: bytes, signature: bytes, padding: ASymPadding) -> bool:
        h = SHA256.new(data)
        try:
            if padding == ASymPadding.PSS:
                pss.new(self._public_key).verify(h, signature)
            elif padding == ASymPadding.PKCShash1v15:
                pkcs1_15.new(self._public_key).verify(h, signature)
            else:
                raise _NotSupportedError(f"Unsupported RSA signing padding: {padding}")
            return True
        except Exception:
            return False


class _DSA_KEYPAIR(_DSA_KEYPAIRTYPE):
    backend = Backend.pycryptodomex
    __concrete__ = True

    def __init__(self, key_size: int, pwd: bytes | str | None = None) -> None:
        self._private_key = DSA.generate(key_size)
        self._public_key = self._private_key.publickey()

    @classmethod
    def decode_private_key(cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding, password: bytes | None = None) -> _ty.Self:
        key = DSA.import_key(data, passphrase=password)
        obj = cls(key.p.bit_length())
        obj._private_key = key
        obj._public_key = key.publickey()
        return obj

    @classmethod
    def decode_public_key(cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> _ty.Self:
        key = DSA.import_key(data)
        obj = cls(2048)
        obj._private_key = None
        obj._public_key = key
        return obj

    def encode_private_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding, password: bytes | None = None) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        fmt = "PEM" if encoding == ASymKeyEncoding.PEM else "DER"
        return self._private_key.export_key(format=fmt, passphrase=password, pkcs8=True)

    def encode_public_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> bytes:
        fmt = "PEM" if encoding == ASymKeyEncoding.PEM else "DER"
        return self._public_key.export_key(format=fmt)

    def sign(self, data: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        signer = DSS.new(self._private_key, "fips-186-3")
        return signer.sign(SHA256.new(data))

    def sign_verify(self, data: bytes, signature: bytes, padding: ASymPadding) -> bool:
        verifier = DSS.new(self._public_key, "fips-186-3")
        try:
            verifier.verify(SHA256.new(data), signature)
            return True
        except Exception:
            return False


_CURVE_MAP = {
    ECCCurve.SECP192R1: "P-192",
    ECCCurve.SECP224R1: "P-224",
    ECCCurve.SECP256R1: "P-256",
    ECCCurve.SECP384R1: "P-384",
    ECCCurve.SECP521R1: "P-521",
    ECCCurve.SECP256K1: "secp256k1",
}


class _ECC_KEYPAIR(_ECC_KEYPAIRTYPE):
    backend = Backend.pycryptodomex
    __concrete__ = True

    def __init__(self, ecc_type: ECCType = ECCType.ECDSA, ecc_curve: ECCCurve | None = ECCCurve.SECP256R1, pwd: bytes | str | None = None) -> None:
        self._ecc_type = ecc_type
        self._ecc_curve = ecc_curve

        if ecc_type == ECCType.ECDSA:
            curve = _CURVE_MAP.get(ecc_curve or ECCCurve.SECP256R1)
            if curve is None:
                raise _NotSupportedError(f"Unsupported ECC curve for pycryptodomex: {ecc_curve}")
            self._private_key = ECC.generate(curve=curve)
        elif ecc_type == ECCType.Ed25519:
            self._private_key = ECC.generate(curve="Ed25519")
        elif ecc_type == ECCType.Ed448:
            self._private_key = ECC.generate(curve="Ed448")
        elif ecc_type == ECCType.X25519:
            self._private_key = ECC.generate(curve="Curve25519")
        elif ecc_type == ECCType.X448:
            self._private_key = ECC.generate(curve="Curve448")
        else:
            raise _NotSupportedError(f"Unsupported ECC type {ecc_type}")
        self._public_key = self._private_key.public_key()

    @classmethod
    def decode_private_key(cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding, password: bytes | None = None) -> _ty.Self:
        key = ECC.import_key(data, passphrase=password)
        curve_name = key.curve
        obj = cls(ECCType.ECDSA, ECCCurve.SECP256R1)
        obj._private_key = key
        obj._public_key = key.public_key()
        if curve_name in ("Ed25519",):
            obj._ecc_type = ECCType.Ed25519
            obj._ecc_curve = None
        elif curve_name in ("Ed448",):
            obj._ecc_type = ECCType.Ed448
            obj._ecc_curve = None
        elif curve_name in ("Curve25519",):
            obj._ecc_type = ECCType.X25519
            obj._ecc_curve = None
        elif curve_name in ("Curve448",):
            obj._ecc_type = ECCType.X448
            obj._ecc_curve = None
        else:
            obj._ecc_type = ECCType.ECDSA
            obj._ecc_curve = next((k for k, v in _CURVE_MAP.items() if v == curve_name), ECCCurve.SECP256R1)
        return obj

    @classmethod
    def decode_public_key(cls, data: bytes, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> _ty.Self:
        key = ECC.import_key(data)
        obj = cls(ECCType.ECDSA, ECCCurve.SECP256R1)
        obj._private_key = None
        obj._public_key = key
        curve_name = key.curve
        if curve_name == "Ed25519":
            obj._ecc_type = ECCType.Ed25519
            obj._ecc_curve = None
        elif curve_name == "Ed448":
            obj._ecc_type = ECCType.Ed448
            obj._ecc_curve = None
        elif curve_name == "Curve25519":
            obj._ecc_type = ECCType.X25519
            obj._ecc_curve = None
        elif curve_name == "Curve448":
            obj._ecc_type = ECCType.X448
            obj._ecc_curve = None
        else:
            obj._ecc_type = ECCType.ECDSA
            obj._ecc_curve = next((k for k, v in _CURVE_MAP.items() if v == curve_name), ECCCurve.SECP256R1)
        return obj

    def encode_private_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding, password: bytes | None = None) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        fmt = "PEM" if encoding == ASymKeyEncoding.PEM else "DER"
        if encoding == ASymKeyEncoding.RAW:
            raise _NotSupportedError("RAW private ECC export is unsupported")
        return self._private_key.export_key(format=fmt, passphrase=password)

    def encode_public_key(self, format_: ASymKeyFormat, encoding: ASymKeyEncoding) -> bytes:
        if encoding == ASymKeyEncoding.RAW:
            return self._public_key.export_key(format="raw")
        if format_ == ASymKeyFormat.OPENSSH:
            return self._public_key.export_key(format="OpenSSH")
        fmt = "PEM" if encoding == ASymKeyEncoding.PEM else "DER"
        return self._public_key.export_key(format=fmt)

    def sign(self, data: bytes, padding: ASymPadding) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        if self._ecc_type == ECCType.ECDSA:
            signer = DSS.new(self._private_key, "fips-186-3")
            return signer.sign(SHA256.new(data))
        if self._ecc_type in (ECCType.Ed25519, ECCType.Ed448):
            signer = eddsa.new(self._private_key, mode="rfc8032")
            return signer.sign(data)
        raise _NotSupportedError(f"{self._ecc_type} does not support signing")

    def sign_verify(self, data: bytes, signature: bytes, padding: ASymPadding) -> bool:
        try:
            if self._ecc_type == ECCType.ECDSA:
                verifier = DSS.new(self._public_key, "fips-186-3")
                verifier.verify(SHA256.new(data), signature)
                return True
            if self._ecc_type in (ECCType.Ed25519, ECCType.Ed448):
                verifier = eddsa.new(self._public_key, mode="rfc8032")
                verifier.verify(data, signature)
                return True
            raise _NotSupportedError(f"{self._ecc_type} does not support signature verification")
        except Exception:
            return False

    def key_exchange(self, peer_public_key: _ty.Self) -> bytes:
        if self._private_key is None:
            raise ValueError("No private key present")
        if self._ecc_type not in (ECCType.ECDSA, ECCType.X25519, ECCType.X448):
            raise _NotSupportedError(f"{self._ecc_type} does not support key exchange")
        try:
            return DH.key_agreement(static_priv=self._private_key, static_pub=peer_public_key._public_key, kdf=lambda x: x)
        except Exception as exc:
            raise _NotSupportedError("ECC key exchange not available for this key type in pycryptodomex") from exc



from __future__ import annotations

import random
import string

import pytest

from ..passwords import (
    PasswordFilter,
    PasswordGenerator,
    SecurePasswordGenerator,
    SimplePasswordGenerator,
)
from ..rand import Random


def test_password_filter_classifies_and_applies_extra_chars() -> None:
    filt = PasswordFilter(exclude_chars="a1!", extra_chars="Z9!ö")

    assert filt.extra_chars_dict == {
        "letters": "Z",
        "numbers": "9",
        "punctuations": "!",
        "unicode": "ö",
    }

    # excludes then adds set-specific extra chars
    assert filt.apply("ab1!", "letters") == "bZ"
    assert filt.apply("12", "numbers") == "29"
    assert filt.apply("!?", "punctuations") == "?!"


def test_password_filter_exclude_similar_and_will_filter() -> None:
    filt = PasswordFilter(exclude_chars="x", exclude_similar=True)

    assert filt.apply("il1Lo0OxY", "letters") == "Y"
    assert filt.filter_word("lilox") == ""
    assert filt.will_filter("l") is True
    assert filt.will_filter("A") is False


def test_simple_password_generator_pattern_and_complex_password() -> None:
    original_rng = SimplePasswordGenerator._rng
    try:
        SimplePasswordGenerator._rng = random.Random(123)
        pat = SimplePasswordGenerator.generate_pattern_password("Xx9-9")
        complex_pw = SimplePasswordGenerator.generate_complex_password(length=14)
    finally:
        SimplePasswordGenerator._rng = original_rng

    assert len(pat) == 5
    assert pat[0] in string.ascii_uppercase
    assert pat[1] in string.ascii_lowercase
    assert pat[2] in string.digits
    assert pat[3] == "-"
    assert pat[4] in string.digits

    assert len(complex_pw) == 14
    assert any(c.isupper() for c in complex_pw)
    assert any(c.islower() for c in complex_pw)
    assert any(c.isdigit() for c in complex_pw)
    assert any(c in string.punctuation for c in complex_pw)


def test_password_generator_complex_password_requires_min_length() -> None:
    gen = PasswordGenerator(Random(0))
    with pytest.raises(ValueError, match="at least 4"):
        gen.generate_complex_password(length=3)


def test_password_generator_random_and_pattern_lengths() -> None:
    gen = PasswordGenerator(Random(1))

    random_pw = gen.generate_random_password(length=17)
    pattern_pw = gen.generate_pattern_password("Xx9-9")

    assert len(random_pw) == 17
    assert len(pattern_pw) == 5
    assert pattern_pw[0] in string.ascii_uppercase
    assert pattern_pw[1] in string.ascii_lowercase
    assert pattern_pw[2] in string.digits
    assert pattern_pw[3] == "-"
    assert pattern_pw[4] in string.digits


def test_password_generator_sentence_invalid_char_position() -> None:
    gen = PasswordGenerator(Random(2))
    result = gen.generate_sentence_based_password_v3(
        sentence="alpha beta",
        char_position="invalid",  # type: ignore[arg-type]
    )
    assert result == "Invalid char_position."


def test_password_generator_words_and_sentence_return_info() -> None:
    gen = PasswordGenerator(Random(3))

    words_pw, source_sentence = gen.generate_words_based_password_v3(
        sentence="alpha beta gamma",
        shuffle_words=False,
        shuffle_characters=False,
        _return_info=True,
    )
    sentence_pw, sentence_source = gen.generate_sentence_based_password_v3(
        sentence="alpha beta",
        char_position="keep",
        random_case=False,
        extra_char="/",
        num_length=2,
        special_chars_length=1,
        password_format="{words}{extra}{numbers}{special}",
        _return_info=True,
    )

    assert source_sentence == "alpha beta gamma"
    assert words_pw == "alphabetagamma"
    assert sentence_source == "alpha beta"
    assert sentence_pw.startswith("ab/")
    assert len(sentence_pw) == len("ab/") + 2 + 1


def test_password_generator_ratio_v3_respects_length_and_filter() -> None:
    gen = PasswordGenerator(Random(4))
    filt = PasswordFilter(exclude_chars=string.ascii_letters + string.punctuation)

    pw = gen.generate_ratio_based_password_v3(
        length=16,
        letters_ratio=0.6,
        numbers_ratio=0.4,
        punctuations_ratio=0.0,
        unicode_ratio=0.0,
        filter_=filt,
    )

    assert len(pw) == 16
    assert all(c in string.digits for c in pw)


def test_password_generator_reduce_password_length_contract() -> None:
    gen = PasswordGenerator(Random(5))

    assert len(gen.reduce_password("abcdef", by=0)) == 6
    assert len(gen.reduce_password("abcdef", by=2)) == 2
    assert len(gen.reduce_password("abcdef", by="all")) == 1


def test_secure_password_generator_methods_return_expected_shape(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    def _load_words(self: PasswordGenerator) -> None:
        self._add_unique_words(["alpha", "beta", "gamma", "delta"])

    monkeypatch.setattr(PasswordGenerator, "load_def_dict", _load_words)
    monkeypatch.setattr(PasswordGenerator, "load_google_10000_dict", _load_words)
    monkeypatch.setattr(PasswordGenerator, "load_scowl_dict", lambda self, size="50": _load_words(self))
    monkeypatch.setattr(PasswordGenerator, "load_12_dicts", _load_words)

    sec = SecurePasswordGenerator("weak")

    outputs = [
        sec.random(),
        sec.passphrase(),
        sec.pattern(),
        sec.complex(),
        sec.mnemonic(),
        sec.ratio(),
        sec.words(),
        sec.sentence(),
        sec.complex_pattern(),
    ]

    for result in outputs:
        assert set(result.keys()) == {"extra_info", "password"}
        assert isinstance(result["password"], str)
        assert result["password"]
        assert isinstance(result["extra_info"], str)
        assert result["extra_info"]


def test_secure_password_generator_dispatch(monkeypatch: pytest.MonkeyPatch) -> None:
    def _load_words(self: PasswordGenerator) -> None:
        self._add_unique_words(["alpha", "beta", "gamma"])

    monkeypatch.setattr(PasswordGenerator, "load_def_dict", _load_words)
    monkeypatch.setattr(PasswordGenerator, "load_google_10000_dict", _load_words)
    monkeypatch.setattr(PasswordGenerator, "load_scowl_dict", lambda self, size="50": _load_words(self))
    monkeypatch.setattr(PasswordGenerator, "load_12_dicts", _load_words)

    sec = SecurePasswordGenerator("average")
    monkeypatch.setattr(SecurePasswordGenerator, "_exponential", lambda self, *args, **kwargs: 0)

    result = sec.generate_secure_password()

    assert set(result.keys()) == {"extra_info", "password"}
    assert result["password"]



from __future__ import annotations

import pytest

pytest.importorskip("cryptography")

from ..crypto import Backend, set_backend
from ..crypto.algos import Asym, HashAlgorithm, KeyDerivationFunction, Sym
from ..crypto.exceptions import NotSupportedError


@pytest.fixture()
def crypto_backend() -> None:
    set_backend([Backend.cryptography_alpha, Backend.std_lib])


def test_backend_wires_core_primitives(crypto_backend: None) -> None:
    assert Sym.Cipher.AES.key.__concrete__ is True
    assert Sym.Cipher.ChaCha20.key.__concrete__ is True
    assert Asym.Cipher.RSA.keypair.__concrete__ is True
    assert Asym.Cipher.DSA.keypair.__concrete__ is True
    assert Asym.Cipher.ECC.keypair.__concrete__ is True


def test_hash_roundtrips_include_sha1(crypto_backend: None) -> None:
    for hasher in (HashAlgorithm.SHA1, HashAlgorithm.SHA2.SHA256, HashAlgorithm.SHA3.SHA256):
        digest = hasher.hash(b"abc")
        assert hasher.verify(b"abc", digest)


def test_kdf_derive_lengths(crypto_backend: None) -> None:
    salt = b"0123456789abcdef"

    assert len(KeyDerivationFunction.PBKDF2HMAC.derive(b"pw", salt=salt, length=16)) == 16
    assert len(KeyDerivationFunction.Scrypt.derive(b"pw", salt=salt, length=16, n=2**10)) == 16
    assert len(KeyDerivationFunction.HKDF.derive(b"pw", salt=salt, length=16)) == 16
    assert len(KeyDerivationFunction.X963.derive(b"pw", otherinfo=b"ctx", length=16)) == 16
    assert len(KeyDerivationFunction.ConcatKDF.derive(b"pw", otherinfo=b"ctx", length=16)) == 16


def test_argon2_variant_guard(crypto_backend: None) -> None:
    with pytest.raises(ValueError, match="variant"):
        KeyDerivationFunction.ARGON2.derive(
            b"pw",
            salt=b"0123456789abcdef",
            length=16,
            variant="i",
        )


def test_kmac_optional_behavior(crypto_backend: None) -> None:
    try:
        out = KeyDerivationFunction.KMAC128.derive(b"pw", length=16)
        assert isinstance(out, bytes)
        assert len(out) == 16
    except ValueError as exc:
        assert "KMAC128" in str(exc)


def test_aes_modes_roundtrip_and_tamper_detection(crypto_backend: None) -> None:
    key = Sym.Cipher.AES.key.new(128)
    plaintext = b"secret payload"

    for mode in (Sym.Operation.ECB, Sym.Operation.CBC, Sym.Operation.CFB, Sym.Operation.OFB, Sym.Operation.CTR):
        ciphertext = key.encrypt(plaintext, Sym.Padding.PKCS7, mode)
        assert key.decrypt(ciphertext, Sym.Padding.PKCS7, mode) == plaintext

    packed_gcm = key.encrypt(plaintext, Sym.Padding.PKCS7, Sym.Operation.GCM)
    assert key.decrypt(packed_gcm, Sym.Padding.PKCS7, Sym.Operation.GCM) == plaintext

    tampered = bytearray(packed_gcm)
    tampered[-1] ^= 0x01
    with pytest.raises(Exception):
        key.decrypt(bytes(tampered), Sym.Padding.PKCS7, Sym.Operation.GCM)


def test_aes_encodings_roundtrip(crypto_backend: None) -> None:
    key = Sym.Cipher.AES.key.new(256)
    plain = b"roundtrip"

    for enc in (Sym.KeyEncoding.RAW, Sym.KeyEncoding.BASE64, Sym.KeyEncoding.HEX, Sym.KeyEncoding.BASE32):
        encoded = key.encode(enc)
        decoded = Sym.Cipher.AES.key.decode(encoded, enc)
        ct = decoded.encrypt(plain, Sym.Padding.PKCS7, Sym.Operation.ECB)
        assert decoded.decrypt(ct, Sym.Padding.PKCS7, Sym.Operation.ECB) == plain


def test_chacha20_roundtrip_and_rejects_block_mode(crypto_backend: None) -> None:
    key = Sym.Cipher.ChaCha20.key.new()
    plain = b"stream-data"

    ct = key.encrypt(plain, Sym.Padding.PKCS7, Sym.Operation.CTR)
    assert key.decrypt(ct, Sym.Padding.PKCS7, Sym.Operation.CTR) == plain

    with pytest.raises(NotSupportedError):
        key.encrypt(plain, Sym.Padding.PKCS7, Sym.Operation.CBC)


def test_mac_generation_and_verification(crypto_backend: None) -> None:
    key = Sym.Cipher.AES.key.new(128)
    data = b"mac-data"

    hmac_mac = key.generate_mac(data, Sym.MessageAuthenticationCode.HMAC)
    assert key.verify_mac(data, hmac_mac, Sym.MessageAuthenticationCode.HMAC)

    cmac_mac = key.generate_mac(data, Sym.MessageAuthenticationCode.CMAC)
    assert key.verify_mac(data, cmac_mac, Sym.MessageAuthenticationCode.CMAC)


def test_rsa_sign_encrypt_and_serialization(crypto_backend: None) -> None:
    key = Asym.Cipher.RSA.keypair.new(2048)
    data = b"hello"

    signature = key.sign(data, Asym.Padding.PSS)
    assert key.sign_verify(data, signature, Asym.Padding.PSS)

    ciphertext = key.encrypt(data, Asym.Padding.OAEP)
    assert key.decrypt(ciphertext, Asym.Padding.OAEP) == data

    pub = key.encode_public_key(Asym.KeyFormat.PKCS1, Asym.KeyEncoding.PEM)
    pub_only = Asym.Cipher.RSA.keypair.decode_public_key(pub, Asym.KeyFormat.PKCS1, Asym.KeyEncoding.PEM)
    assert pub_only.sign_verify(data, signature, Asym.Padding.PSS)


def test_dsa_sign_and_verify(crypto_backend: None) -> None:
    key = Asym.Cipher.DSA.keypair.new(2048)
    data = b"hello"

    signature = key.sign(data, Asym.Padding.PSS)
    assert key.sign_verify(data, signature, Asym.Padding.PSS)


def test_ecc_ecdsa_and_key_exchange(crypto_backend: None) -> None:
    a = Asym.Cipher.ECC.keypair.ecdsa_key(Asym.Cipher.ECC.Curve.SECP256R1)
    b = Asym.Cipher.ECC.keypair.ecdsa_key(Asym.Cipher.ECC.Curve.SECP256R1)
    data = b"hello"

    signature = a.sign(data, Asym.Padding.PSS)
    assert a.sign_verify(data, signature, Asym.Padding.PSS)

    secret_a = a.key_exchange(b)
    secret_b = b.key_exchange(a)
    assert secret_a == secret_b


def test_ecc_ed25519_and_x25519_capabilities(crypto_backend: None) -> None:
    ed = Asym.Cipher.ECC.keypair.optimized_key(Asym.Cipher.ECC.Type.Ed25519)
    data = b"hello"

    signature = ed.sign(data, Asym.Padding.PSS)
    assert ed.sign_verify(data, signature, Asym.Padding.PSS)
    with pytest.raises(NotSupportedError):
        ed.key_exchange(ed)

    x1 = Asym.Cipher.ECC.keypair.optimized_key(Asym.Cipher.ECC.Type.X25519)
    x2 = Asym.Cipher.ECC.keypair.optimized_key(Asym.Cipher.ECC.Type.X25519)
    assert x1.key_exchange(x2) == x2.key_exchange(x1)


def test_backend_fallback_with_stdlib(crypto_backend: None) -> None:
    digest = HashAlgorithm.SHA2.SHA256.hash(b"hello")
    assert HashAlgorithm.SHA2.SHA256.verify(b"hello", digest)



import bcrypt


def hash_bcrypt(password: bytes) -> bytes:
    if not password:
        raise ValueError("Password must not be empty")
    return bcrypt.hashpw(password, bcrypt.gensalt())


def hash_verify_bcrypt(password: bytes, hash_: bytes) -> bool:
    if not password:
        raise ValueError("Password must not be empty")
    return bcrypt.checkpw(password, hash_)


def derive_bcrypt(
    password: bytes, salt: bytes, rounds: int = 12, length: int = 32
) -> bytes:
    if len(salt) < 16:
        raise ValueError("Salt must be at least 16 bytes for bcrypt.kdf")
    return bcrypt.kdf(
        password=password, salt=salt, desired_key_bytes=length, rounds=rounds
    )



from __future__ import annotations

import hmac
from typing import Callable

from Cryptodome.Hash import (
    BLAKE2b,
    BLAKE2s,
    MD5,
    RIPEMD160,
    SHA1,
    SHA224,
    SHA256,
    SHA384,
    SHA3_224,
    SHA3_256,
    SHA3_384,
    SHA3_512,
    SHA512,
    #SHA512_224,
    #SHA512_256,
    SHAKE128,
    SHAKE256,
)
from Cryptodome.Protocol.KDF import HKDF, PBKDF1, PBKDF2, scrypt

try:
    from Cryptodome.Hash import KMAC128, KMAC256
except Exception:  # pragma: no cover
    KMAC128 = None
    KMAC256 = None

_HASH_FNS: dict[str, Callable[..., bytes]] = {}


def _reg(name: str):
    def _wrap(fn: Callable[..., bytes]):
        _HASH_FNS[name] = fn
        return fn

    return _wrap


@_reg("sha1")
def hash_sha1(data: bytes) -> bytes:
    return SHA1.new(data).digest()


@_reg("sha224")
def hash_sha224(data: bytes) -> bytes:
    return SHA224.new(data).digest()


@_reg("sha256")
def hash_sha256(data: bytes) -> bytes:
    return SHA256.new(data).digest()


@_reg("sha384")
def hash_sha384(data: bytes) -> bytes:
    return SHA384.new(data).digest()


@_reg("sha512")
def hash_sha512(data: bytes) -> bytes:
    return SHA512.new(data).digest()


@_reg("sha512_224")
def hash_sha512_224(data: bytes) -> bytes:
    return SHA512_224.new(data).digest()


@_reg("sha512_256")
def hash_sha512_256(data: bytes) -> bytes:
    return SHA512_256.new(data).digest()


@_reg("sha3_224")
def hash_sha3_224(data: bytes) -> bytes:
    return SHA3_224.new(data).digest()


@_reg("sha3_256")
def hash_sha3_256(data: bytes) -> bytes:
    return SHA3_256.new(data).digest()


@_reg("sha3_384")
def hash_sha3_384(data: bytes) -> bytes:
    return SHA3_384.new(data).digest()


@_reg("sha3_512")
def hash_sha3_512(data: bytes) -> bytes:
    return SHA3_512.new(data).digest()


@_reg("sha3_shake_128")
def hash_sha3_shake_128(data: bytes, length: int = 32) -> bytes:
    return SHAKE128.new(data).read(length)


@_reg("sha3_shake_256")
def hash_sha3_shake_256(data: bytes, length: int = 32) -> bytes:
    return SHAKE256.new(data).read(length)


@_reg("blake2b")
def hash_blake2b(data: bytes, digest_size: int = 64) -> bytes:
    return BLAKE2b.new(data=data, digest_bytes=digest_size).digest()


@_reg("blake2s")
def hash_blake2s(data: bytes, digest_size: int = 32) -> bytes:
    return BLAKE2s.new(data=data, digest_bits=digest_size * 8).digest()


@_reg("md5")
def hash_md5(data: bytes) -> bytes:
    return MD5.new(data).digest()


@_reg("ripemd160")
def hash_ripemd160(data: bytes) -> bytes:
    return RIPEMD160.new(data).digest()


def _verify(name: str, data: bytes, expected: bytes, *args) -> bool:
    fn = _HASH_FNS[name]
    return hmac.compare_digest(fn(data, *args), expected)


def hash_verify_sha1(data: bytes, expected: bytes) -> bool:
    return _verify("sha1", data, expected)


def hash_verify_sha224(data: bytes, expected: bytes) -> bool:
    return _verify("sha224", data, expected)


def hash_verify_sha256(data: bytes, expected: bytes) -> bool:
    return _verify("sha256", data, expected)


def hash_verify_sha384(data: bytes, expected: bytes) -> bool:
    return _verify("sha384", data, expected)


def hash_verify_sha512(data: bytes, expected: bytes) -> bool:
    return _verify("sha512", data, expected)


def hash_verify_sha512_224(data: bytes, expected: bytes) -> bool:
    return _verify("sha512_224", data, expected)


def hash_verify_sha512_256(data: bytes, expected: bytes) -> bool:
    return _verify("sha512_256", data, expected)


def hash_verify_sha3_224(data: bytes, expected: bytes) -> bool:
    return _verify("sha3_224", data, expected)


def hash_verify_sha3_256(data: bytes, expected: bytes) -> bool:
    return _verify("sha3_256", data, expected)


def hash_verify_sha3_384(data: bytes, expected: bytes) -> bool:
    return _verify("sha3_384", data, expected)


def hash_verify_sha3_512(data: bytes, expected: bytes) -> bool:
    return _verify("sha3_512", data, expected)


def hash_verify_sha3_shake_128(data: bytes, expected: bytes, length: int = 32) -> bool:
    return _verify("sha3_shake_128", data, expected, length)


def hash_verify_sha3_shake_256(data: bytes, expected: bytes, length: int = 32) -> bool:
    return _verify("sha3_shake_256", data, expected, length)


def hash_verify_blake2b(data: bytes, expected: bytes, digest_size: int = 64) -> bool:
    return _verify("blake2b", data, expected, digest_size)


def hash_verify_blake2s(data: bytes, expected: bytes, digest_size: int = 32) -> bool:
    return _verify("blake2s", data, expected, digest_size)


def hash_verify_md5(data: bytes, expected: bytes) -> bool:
    return _verify("md5", data, expected)


def hash_verify_ripemd160(data: bytes, expected: bytes) -> bool:
    return _verify("ripemd160", data, expected)


_HASH_MODS = {
    "sha1": SHA1,
    "sha224": SHA224,
    "sha256": SHA256,
    "sha384": SHA384,
    "sha512": SHA512,
#    "sha512_224": SHA512_224,
#    "sha512_256": SHA512_256,
    "sha3_224": SHA3_224,
    "sha3_256": SHA3_256,
    "sha3_384": SHA3_384,
    "sha3_512": SHA3_512,
    "md5": MD5,
    "ripemd160": RIPEMD160,
}


def _get_hash_mod(name: str):
    mod = _HASH_MODS.get(name)
    if mod is None:
        raise ValueError(f"Unsupported hash algorithm '{name}'")
    return mod


def derive_pbkdf2hmac(password: bytes, salt: bytes, length: int, iterations: int, algorithm: str) -> bytes:
    return PBKDF2(password, salt, dkLen=length, count=iterations, hmac_hash_module=_get_hash_mod(algorithm))


def derive_pbkdf1(password: bytes, salt: bytes, length: int, iterations: int, algorithm: str) -> bytes:
    return PBKDF1(password, salt, dkLen=length, count=iterations, hashAlgo=_get_hash_mod(algorithm))


def derive_scrypt(password: bytes, salt: bytes, length: int, n: int, r: int, p: int) -> bytes:
    return scrypt(password, salt=salt, key_len=length, N=n, r=r, p=p)


def derive_hkdf(password: bytes, salt: bytes, info: bytes, length: int, algorithm: str) -> bytes:
    return HKDF(password, key_len=length, salt=salt, hashmod=_get_hash_mod(algorithm), context=info)


def derive_x963(password: bytes, length: int, otherinfo: bytes, algorithm: str) -> bytes:
    # ANSI X9.63 style KDF approximation using counter || Z || sharedinfo.
    return derive_concatkdf(password, length, otherinfo, algorithm)


def derive_concatkdf(password: bytes, length: int, otherinfo: bytes, algorithm: str) -> bytes:
    hash_mod = _get_hash_mod(algorithm)
    out = b""
    counter = 1
    while len(out) < length:
        h = hash_mod.new()
        h.update(counter.to_bytes(4, "big"))
        h.update(password)
        h.update(otherinfo)
        out += h.digest()
        counter += 1
    return out[:length]


def derive_kmac128(password: bytes, length: int, key: bytes = b"", customization: bytes = b"") -> bytes:
    if KMAC128 is None:
        raise ValueError("KMAC128 is unavailable in installed pycryptodomex")
    kmac = KMAC128.new(key=key, mac_len=length, custom=customization)
    kmac.update(password)
    return kmac.digest()


def derive_kmac256(password: bytes, length: int, key: bytes = b"", customization: bytes = b"") -> bytes:
    if KMAC256 is None:
        raise ValueError("KMAC256 is unavailable in installed pycryptodomex")
    kmac = KMAC256.new(key=key, mac_len=length, custom=customization)
    kmac.update(password)
    return kmac.digest()
